# Copyright 2024 The Treescope Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Single-purpose ndarray visualizer for Python in vanilla JavaScript.

Designed to quickly visualize the contents of arbitrarily-high-dimensional
NDArrays, to allow them to be visualized by default instead of requiring lots
of manual effort.
"""

from __future__ import annotations

import collections
import itertools
import json
from typing import Any, Literal, Sequence

import numpy as np
from treescope import context
from treescope import dtype_util
from treescope import ndarray_adapters
from treescope import type_registries
from treescope._internal import arrayviz_impl
from treescope._internal import figures_impl
from treescope._internal import html_escaping
from treescope._internal.parts import basic_parts
from treescope._internal.parts import common_styles

AxisName = Any

PositionalAxisInfo = ndarray_adapters.PositionalAxisInfo
NamedPositionlessAxisInfo = ndarray_adapters.NamedPositionlessAxisInfo
NamedPositionalAxisInfo = ndarray_adapters.NamedPositionalAxisInfo
AxisInfo = ndarray_adapters.AxisInfo

ArrayInRegistry = Any


default_sequential_colormap: context.ContextualValue[
    list[tuple[int, int, int]]
] = context.ContextualValue(
    module=__name__,
    qualname="default_sequential_colormap",
    # Matplotlib Viridis_20
    initial_value=[
        (68, 1, 84),
        (72, 20, 103),
        (72, 38, 119),
        (69, 55, 129),
        (63, 71, 136),
        (57, 85, 140),
        (50, 100, 142),
        (45, 113, 142),
        (40, 125, 142),
        (35, 138, 141),
        (31, 150, 139),
        (32, 163, 134),
        (41, 175, 127),
        (59, 187, 117),
        (86, 198, 103),
        (115, 208, 86),
        (149, 216, 64),
        (184, 222, 41),
        (221, 227, 24),
        (253, 231, 37),
    ],
)

default_diverging_colormap: context.ContextualValue[
    list[tuple[int, int, int]]
] = context.ContextualValue(
    module=__name__,
    qualname="default_diverging_colormap",
    # cmocean Balance_19_r[1:-1]
    initial_value=[
        (96, 14, 34),
        (134, 14, 41),
        (167, 36, 36),
        (186, 72, 46),
        (198, 107, 77),
        (208, 139, 115),
        (218, 171, 155),
        (228, 203, 196),
        (241, 236, 235),
        (202, 212, 216),
        (161, 190, 200),
        (117, 170, 190),
        (75, 148, 186),
        (38, 123, 186),
        (12, 94, 190),
        (41, 66, 162),
        (37, 47, 111),
    ],
)


def render_array(
    array: ArrayInRegistry,
    *,
    columns: Sequence[AxisName | int] = (),
    rows: Sequence[AxisName | int] = (),
    sliders: Sequence[AxisName | int] = (),
    valid_mask: Any | None = None,
    continuous: bool | Literal["auto"] = "auto",
    around_zero: bool | Literal["auto"] = "auto",
    vmax: float | None = None,
    vmin: float | None = None,
    trim_outliers: bool = True,
    dynamic_colormap: bool | Literal["auto"] = "auto",
    colormap: list[tuple[int, int, int]] | None = None,
    truncate: bool = False,
    maximum_size: int = 10_000,
    cutoff_size_per_axis: int = 512,
    minimum_edge_items: int = 5,
    axis_item_labels: dict[AxisName | int, list[str]] | None = None,
    value_item_labels: dict[int, str] | None = None,
    axis_labels: dict[AxisName | int, str] | None = None,
    pixels_per_cell: int | float = 7,
) -> figures_impl.TreescopeFigure:
  """Renders an array (positional or named) to a displayable HTML object.

  Each element of the array is rendered to a fixed-size square, with its
  position determined based on its index, and with each level of x and y axis
  represented by a "faceted" plot.

  Out-of-bounds or otherwise unusual data is rendered with an annotation:

  * "X" means a value was NaN (for continuous data) or went out-of-bounds for
    the integer palette (for discrete data).

  * "I" or "-I" means a value was infinity or negative infinity.

  * "+" or "-" means a value was finite but went outside the bounds of the
    colormap (e.g. it was larger than ``vmax`` or smaller than ``vmin``). By
    default this applies to values more than 3 standard deviations outside the
    mean.

  * Four light dots on grey means a value was masked out by ``valid_mask``, or
    truncated due to the maximum size or axis cutoff thresholds.

  By default, this method automatically chooses a color rendering strategy based
  on the arguments:

  * If an explicit colormap is provided:

    * If ``continuous`` is True, the provided colors are interpreted as color
      stops and interpolated between.

    * If ``continuous`` is False, the provided colors are interpreted as an
      indexed color palette, and each index of the palette is used to render
      the corresponding integer, starting from zero.

  * Otherwise:

    * If ``continuous`` is True:

      * If ``around_zero`` is True, uses the diverging colormap
        `default_diverging_colormap`. The initial value of this is a truncated
        version of the perceptually-uniform "Balance" colormap from cmocean,
        with blue for positive numbers and red for negative ones.

      * If ``around_zero`` is False, uses the sequential colormap
        `default_sequential_colormap`.The initial value of this is the
        perceptually-uniform "Viridis" colormap from matplotlib.

    * If ``continuous`` is False, uses a pattern-based "digitbox" rendering
      strategy to render integers up to 9,999,999 as nested squares, with one
      square per integer digit and digit colors drawn from the D3 Category20
      colormap.

  Args:
      array: The array to render. The type of this array must be registered in
        the `type_registries.NDARRAY_ADAPTER_REGISTRY`.
      columns: Sequence of axis names or positional axis indices that should be
        placed on the x axis, from innermost to outermost. If not provided,
        inferred automatically.
      rows: Sequence of axis names or positional axis indices that should be
        placed on the y axis, from innermost to outermost. If not provided,
        inferred automatically.
      sliders: Sequence of axis names or positional axis indices for which we
        should show only a single slice at a time, with the index determined
        with a slider.
      valid_mask: Optionally, a boolean array with the same shape (and, if
        applicable, axis names) as `array`, which is True for the locations that
        we should actually render, and False for locations that do not have
        valid array data.
      continuous: Whether to interpret this array as numbers along the real
        line, and visualize using an interpolated colormap. If "auto", inferred
        from the dtype of `array`.
      around_zero: Whether the array data should be rendered symmetrically
        around zero using a diverging colormap, scaled based on the absolute
        magnitude of the inputs, instead of rescaled to be between the min and
        max of the data. If "auto", treated as True unless both `vmin` and
        `vmax` are set to incompatible values.
      vmax: Largest value represented in the colormap. If omitted and
        around_zero is True, inferred as ``max(abs(array))`` or as ``-vmin``. If
        omitted and around_zero is False, inferred as ``max(array)``.
      vmin: Smallest value represented in the colormap. If omitted and
        around_zero is True, inferred as ``-max(abs(array))`` or as ``-vmax``.
        If omitted and around_zero is False, inferred as ``min(array)``.
      trim_outliers: Whether to try to trim outliers when inferring ``vmin`` and
        ``vmax``. If True, clips them to 3 standard deviations away from the
        mean (or 3 sqrt-second-moments around zero) if they would otherwise
        exceed it.
      dynamic_colormap: Whether to dynamically adjust the colormap based on
        mouse hover. Requires a continuous colormap, and ``around_zero=True``.
        If "auto", will be enabled for continuous arrays if ``around_zero`` is
        True and neither ``vmin`` nor ``vmax`` are provided.
      colormap: An optional explicit colormap to use, represented as a list of
        ``(r,g,b)`` tuples, where each channel is between 0 and 255. A good
        place to get colormaps is the ``palettable`` package, e.g. you can pass
        something like ``palettable.matplotlib.Inferno_20.colors``.
      truncate: Whether or not to truncate the array to a smaller size before
        rendering.
      maximum_size: Maximum number of elements of an array to show. Arrays
        larger than this will be truncated along one or more axes. Ignored
        unless ``truncate`` is True.
      cutoff_size_per_axis: Maximum number of elements of each individual axis
        to show without truncation. Any axis longer than this will be truncated,
        with their visual size increasing logarithmically with the true axis
        size beyond this point. Ignored unless ``truncate`` is True.
      minimum_edge_items: How many values to keep along each axis for truncated
        arrays. We may keep more than this up to the budget of maximum_size.
        Ignored unless ``truncate`` is True.
      axis_item_labels: An optional mapping from axis names/positions to a list
        of strings, of the same length as the axis length, giving a label to
        each item along that axis. For instance, this could be the token string
        corresponding to each position along a sequence axis, or the class label
        corresponding to each category across a classifier's output axis. This
        is shown in the tooltip when hovering over a pixel, and shown below the
        array when a pixel is clicked on. For convenience, names in this
        dictionary that don't match any axes in the input are simply ignored, so
        that you can pass the same labels while rendering arrays that may not
        have the same axis names.
      value_item_labels: For categorical data, an optional mapping from each
        value to a string. For instance, this could be the token value
        corresponding to each token ID in a sequence of tokens.
      axis_labels: Optional mapping from axis names / indices to the labels we
        should use for that axis. If not provided, we label the named axes with
        their names and the positional axes with "axis {i}", and also add th
        axis size.
      pixels_per_cell: Size of each rendered array element in pixels, between 1
        and 21 inclusive. This controls the zoom level of the rendering. Array
        elements are always drawn at 7 pixels per cell and then rescaled, so
        out-of-bounds annotations and "digitbox" integer value patterns may not
        display correctly at fewer than 7 pixels per cell.

  Returns:
    An object which can be rendered in an IPython notebook, containing the
    HTML source of an arrayviz rendering.
  """
  # Retrieve the adapter for this array, which we will use to construct
  # the rendering.
  type_registries.update_registries_for_imports()
  adapter = type_registries.lookup_ndarray_adapter(array)
  if adapter is None:
    raise TypeError(
        f"Cannot render array with unrecognized type {type(array)} (not found"
        " in array adapter registry)"
    )
  if not 1 <= pixels_per_cell <= 21:
    raise ValueError(
        "pixels_per_cell must be between 1 and 21 inclusive, got"
        f" {pixels_per_cell}"
    )

  # Extract information about axis names, indices, and sizes.
  array_axis_info = adapter.get_axis_info_for_array_data(array)

  data_axis_from_axis_info = {
      info: axis for axis, info in enumerate(array_axis_info)
  }
  assert len(data_axis_from_axis_info) == len(array_axis_info)

  info_by_name_or_position = {}
  for info in array_axis_info:
    if isinstance(info, NamedPositionalAxisInfo):
      info_by_name_or_position[info.axis_name] = info
      info_by_name_or_position[info.axis_logical_index] = info
    elif isinstance(info, PositionalAxisInfo):
      info_by_name_or_position[info.axis_logical_index] = info
    elif isinstance(info, NamedPositionlessAxisInfo):
      info_by_name_or_position[info.axis_name] = info
    else:
      raise ValueError(f"Unrecognized axis info {type(info)}")

  row_infos = [info_by_name_or_position[spec] for spec in rows]
  column_infos = [info_by_name_or_position[spec] for spec in columns]
  slider_infos = [info_by_name_or_position[spec] for spec in sliders]

  unassigned_axes = set(array_axis_info)
  seen_axes = set()
  for axis_info in itertools.chain(row_infos, column_infos, slider_infos):
    if axis_info in seen_axes:
      raise ValueError(
          f"Axis {axis_info} appeared multiple times in rows/columns/sliders"
          " specifications. Each axis must be assigned to at most one"
          " location."
      )
    seen_axes.add(axis_info)
    unassigned_axes.remove(axis_info)

  if truncate:
    # Infer a good truncated shape for this array.
    edge_items_per_axis = arrayviz_impl.infer_balanced_truncation(
        tuple(info.size for info in array_axis_info),
        maximum_size=maximum_size,
        cutoff_size_per_axis=cutoff_size_per_axis,
        minimum_edge_items=minimum_edge_items,
    )
  else:
    edge_items_per_axis = (None,) * len(array_axis_info)

  # Obtain truncated array and mask data from the adapter.
  truncated_array_data, truncated_mask_data = (
      adapter.get_array_data_with_truncation(
          array=array,
          mask=valid_mask,
          edge_items_per_axis=edge_items_per_axis,
      )
  )

  # Step 5: Figure out which axes to render as rows, columns, and sliders and
  # in which order. We start with the explicitly-requested axes, then add more
  # axes to the rows and columns until we've assigned all of them, trying to
  # balance rows and columns.

  row_infos, column_infos = arrayviz_impl.infer_rows_and_columns(
      all_axes=[ax for ax in array_axis_info if ax not in slider_infos],
      known_rows=row_infos,
      known_columns=column_infos,
      edge_items_per_axis=edge_items_per_axis,
  )

  return figures_impl.TreescopeFigure(
      _render_pretruncated(
          array_axis_info=array_axis_info,
          row_infos=row_infos,
          column_infos=column_infos,
          slider_infos=slider_infos,
          truncated_array_data=truncated_array_data,
          truncated_mask_data=truncated_mask_data,
          edge_items_per_axis=edge_items_per_axis,
          continuous=continuous,
          around_zero=around_zero,
          vmax=vmax,
          vmin=vmin,
          trim_outliers=trim_outliers,
          dynamic_colormap=dynamic_colormap,
          colormap=colormap,
          axis_item_labels=axis_item_labels,
          value_item_labels=value_item_labels,
          axis_labels=axis_labels,
          pixels_per_cell=pixels_per_cell,
      )
  )


def _render_pretruncated(
    *,
    array_axis_info: Sequence[AxisInfo],
    row_infos: Sequence[AxisInfo],
    column_infos: Sequence[AxisInfo],
    slider_infos: Sequence[AxisInfo],
    truncated_array_data: np.ndarray,
    truncated_mask_data: np.ndarray,
    edge_items_per_axis: Sequence[int | None],
    continuous: bool | Literal["auto"],
    around_zero: bool | Literal["auto"],
    vmax: float | None,
    vmin: float | None,
    trim_outliers: bool,
    dynamic_colormap: bool | Literal["auto"],
    colormap: list[tuple[int, int, int]] | None,
    axis_item_labels: dict[AxisName | int, list[str]] | None,
    value_item_labels: dict[int, str] | None,
    axis_labels: dict[AxisName | int, str] | None,
    pixels_per_cell: int | float = 7,
) -> arrayviz_impl.ArrayvizRendering:
  """Internal helper to render an array that has already been truncated."""
  if axis_item_labels is None:
    axis_item_labels = {}

  if value_item_labels is None:
    value_item_labels = {}

  if axis_labels is None:
    axis_labels = {}

  data_axis_from_axis_info = {
      info: axis for axis, info in enumerate(array_axis_info)
  }
  assert len(data_axis_from_axis_info) == len(array_axis_info)

  has_name_only = False
  positional_count = 0

  info_by_name_or_position = {}
  for info in array_axis_info:
    if isinstance(info, NamedPositionalAxisInfo):
      info_by_name_or_position[info.axis_name] = info
      info_by_name_or_position[info.axis_logical_index] = info
      positional_count += 1
    elif isinstance(info, PositionalAxisInfo):
      info_by_name_or_position[info.axis_logical_index] = info
      positional_count += 1
    elif isinstance(info, NamedPositionlessAxisInfo):
      info_by_name_or_position[info.axis_name] = info
      has_name_only = True
    else:
      raise ValueError(f"Unrecognized axis info {type(info)}")

  axis_labels_by_info = {
      info_by_name_or_position[orig_key]: value
      for orig_key, value in axis_labels.items()
  }
  axis_item_labels_by_info = {
      info_by_name_or_position[orig_key]: value
      for orig_key, value in axis_item_labels.items()
  }

  skip_start_indices = [
      edge_items if edge_items is not None else axis_info.size
      for edge_items, axis_info in zip(edge_items_per_axis, array_axis_info)
  ]
  skip_end_indices = [
      axis_info.size - edge_items if edge_items is not None else axis_info.size
      for edge_items, axis_info in zip(edge_items_per_axis, array_axis_info)
  ]

  # Convert the axis names into indices into our data array.
  column_data_axes = [
      data_axis_from_axis_info[orig_axis] for orig_axis in column_infos
  ]
  row_data_axes = [
      data_axis_from_axis_info[orig_axis] for orig_axis in row_infos
  ]
  slider_data_axes = [
      data_axis_from_axis_info[orig_axis] for orig_axis in slider_infos
  ]

  # Step 6: Figure out how to render the labels and indices of each axis.
  # We render indices using a small interpreted format language that can be
  # serialized to JSON and interpreted in JavaScript.
  data_axis_labels = {}
  formatting_instructions = []
  formatting_instructions.append({"type": "literal", "value": "array"})

  axis_label_instructions = []

  if has_name_only:
    formatting_instructions.append({"type": "literal", "value": "[{"})

    first = True
    for data_axis, axis_info in enumerate(array_axis_info):
      if not isinstance(axis_info, NamedPositionlessAxisInfo):
        continue

      if first:
        formatting_instructions.append(
            {"type": "literal", "value": f"{repr(axis_info.axis_name)}:"}
        )
        first = False
      else:
        formatting_instructions.append(
            {"type": "literal", "value": f", {repr(axis_info.axis_name)}:"}
        )

      formatting_instructions.append({
          "type": "index",
          "axis": f"a{data_axis}",
          "skip_start": skip_start_indices[data_axis],
          "skip_end": skip_end_indices[data_axis],
      })

      if axis_info in axis_labels_by_info:
        data_axis_labels[data_axis] = axis_labels_by_info[axis_info]
        label_name = f"{axis_labels_by_info[axis_info]} ({axis_info.axis_name})"
      elif axis_info in slider_infos:
        label_name = f"{str(axis_info.axis_name)}"
        data_axis_labels[data_axis] = label_name
      else:
        label_name = f"{str(axis_info.axis_name)}"
        data_axis_labels[data_axis] = f"{label_name}: {axis_info.size}"

      if axis_info in axis_item_labels_by_info:
        axis_label_instructions.extend([
            {"type": "literal", "value": f"\n{label_name} @ "},
            {
                "type": "index",
                "axis": f"a{data_axis}",
                "skip_start": skip_start_indices[data_axis],
                "skip_end": skip_end_indices[data_axis],
            },
            {"type": "literal", "value": ": "},
            {
                "type": "axis_lookup",
                "axis": f"a{data_axis}",
                "skip_start": skip_start_indices[data_axis],
                "skip_end": skip_end_indices[data_axis],
                "lookup_table": axis_item_labels_by_info[axis_info],
            },
        ])

    formatting_instructions.append({"type": "literal", "value": "}]"})

  if positional_count:
    formatting_instructions.append({"type": "literal", "value": "["})
    for logical_index in range(positional_count):
      axis_info = info_by_name_or_position[logical_index]
      assert isinstance(axis_info, PositionalAxisInfo | NamedPositionalAxisInfo)
      assert axis_info.axis_logical_index == logical_index
      data_axis = data_axis_from_axis_info[axis_info]
      if logical_index > 0:
        formatting_instructions.append({"type": "literal", "value": ", "})
      formatting_instructions.append({
          "type": "index",
          "axis": f"a{data_axis}",
          "skip_start": skip_start_indices[data_axis],
          "skip_end": skip_end_indices[data_axis],
      })

      if axis_info in axis_labels_by_info:
        data_axis_labels[data_axis] = axis_labels_by_info[axis_info]
        label_name = f"{axis_labels_by_info[axis_info]} (axis {logical_index})"
      else:
        if isinstance(axis_info, NamedPositionalAxisInfo):
          label_name = f"{axis_info.axis_name} (axis {logical_index})"
        else:
          label_name = f"axis {logical_index}"
        if axis_info in slider_infos:
          data_axis_labels[data_axis] = label_name
        else:
          data_axis_labels[data_axis] = f"{label_name}: {axis_info.size}"

      if axis_info in axis_item_labels_by_info:
        axis_label_instructions.extend([
            {"type": "literal", "value": f"\n{label_name} @ "},
            {
                "type": "index",
                "axis": f"a{data_axis}",
                "skip_start": skip_start_indices[data_axis],
                "skip_end": skip_end_indices[data_axis],
            },
            {"type": "literal", "value": ": "},
            {
                "type": "axis_lookup",
                "axis": f"a{data_axis}",
                "skip_start": skip_start_indices[data_axis],
                "skip_end": skip_end_indices[data_axis],
                "lookup_table": axis_item_labels_by_info[axis_info],
            },
        ])

    formatting_instructions.append({"type": "literal", "value": "]"})

  formatting_instructions.append({"type": "literal", "value": "\n  = "})
  formatting_instructions.append({"type": "value"})

  # Step 7: Infer the colormap and rendering strategy.

  # Figure out whether the array is continuous.
  inferred_continuous = dtype_util.is_floating_dtype(truncated_array_data.dtype)
  if continuous == "auto":
    continuous = inferred_continuous
  elif not continuous and inferred_continuous:
    raise ValueError(
        "Cannot use continuous=False when rendering a float array; explicitly"
        " cast it to an integer array first."
    )

  if inferred_continuous:
    # Cast to float32 to ensure we can easily manipulate the truncated data.
    truncated_array_data = truncated_array_data.astype(np.float32)

  if value_item_labels and not continuous:
    formatting_instructions.append({"type": "literal", "value": "  # "})
    formatting_instructions.append(
        {"type": "value_lookup", "lookup_table": value_item_labels}
    )

  formatting_instructions.extend(axis_label_instructions)

  # Figure out centering.
  definitely_not_around_zero = (
      vmin is not None and vmax is not None and vmin != -vmax  # pylint: disable=invalid-unary-operand-type
  )
  if around_zero == "auto":
    around_zero = not definitely_not_around_zero
  elif around_zero and definitely_not_around_zero:
    raise ValueError(
        "Cannot use around_zero=True while also specifying both vmin and vmax"
    )

  # Check whether we should dynamically adjust the colormap.
  if dynamic_colormap == "auto":
    dynamic_colormap = (
        continuous and around_zero and vmin is None and vmax is None
    )

  if dynamic_colormap:
    if not continuous:
      raise ValueError(
          "Cannot use dynamic_colormap with a non-continuous colormap."
      )
    if not around_zero:
      raise ValueError("Cannot use dynamic_colormap without around_zero.")

    raw_min_abs, raw_max_abs = arrayviz_impl.infer_abs_min_max(
        truncated_array_data, truncated_mask_data
    )
    raw_min_abs = float(raw_min_abs)
    raw_max_abs = float(raw_max_abs)
  else:
    raw_min_abs = None
    raw_max_abs = None

  # Infer concrete `vmin` and `vmax`.
  if continuous and (vmin is None or vmax is None):
    vmin, vmax = arrayviz_impl.infer_vmin_vmax(
        array=truncated_array_data,
        mask=truncated_mask_data,
        vmin=vmin,
        vmax=vmax,
        around_zero=around_zero,
        trim_outliers=trim_outliers,
    )
    vmin = float(vmin)
    vmax = float(vmax)

  # Figure out which colormap and rendering strategy to use.
  if colormap is None:
    if continuous:
      colormap_type = "continuous"
      if around_zero:
        colormap_data = default_diverging_colormap.get()
      else:
        colormap_data = default_sequential_colormap.get()

    else:
      colormap_type = "digitbox"
      colormap_data = []

  elif continuous:
    colormap_data = colormap
    colormap_type = "continuous"

  else:
    colormap_data = colormap
    colormap_type = "palette_index"

  # Make a title for it
  info_parts = []
  if dynamic_colormap:
    info_parts.append("Dynamic colormap (click to adjust).")
  elif continuous:
    info_parts.append(f"Linear colormap from {vmin:.6g} to {vmax:.6g}.")
  elif colormap is not None:
    info_parts.append("Indexed colors from a color list.")
  else:
    info_parts.append("Showing integer digits as nested squares.")

  info_parts.append(" Hover/click for array data.")

  # Step 8: Render it!
  html_src = arrayviz_impl.render_array_data_to_html(
      array_data=truncated_array_data,
      valid_mask=truncated_mask_data,
      column_axes=column_data_axes,
      row_axes=row_data_axes,
      slider_axes=slider_data_axes,
      axis_labels=[
          data_axis_labels[i] for i in range(truncated_array_data.ndim)
      ],
      vmin=vmin,
      vmax=vmax,
      cmap_type=colormap_type,
      cmap_data=colormap_data,
      info="".join(info_parts),
      formatting_instructions=formatting_instructions,
      dynamic_continuous_cmap=dynamic_colormap,
      raw_min_abs=raw_min_abs,
      raw_max_abs=raw_max_abs,
      pixels_per_cell=pixels_per_cell,
  )
  return arrayviz_impl.ArrayvizRendering(html_src)


def render_sharding_info(
    array_axis_info: Sequence[AxisInfo],
    sharding_info: ndarray_adapters.ShardingInfo,
    rows: Sequence[int | AxisName] = (),
    columns: Sequence[int | AxisName] = (),
) -> figures_impl.TreescopeFigure:
  """Renders the sharding of an array.

  This is a helper function for rendering array shardings. It can be used either
  to render the sharding of an actual array or of a hypothetical array of a
  given shape and sharding.

  Args:
    array_axis_info: Axis info for each axis of the array data.
    sharding_info: Sharding info for the array, as produced by a NDArrayAdapter.
    rows: Optional explicit ordering of rows in the visualization.
    columns: Optional explicit ordering of columns in the visualization.

  Returns:
    A rendering of the sharding, which re-uses the digitbox rendering mode to
    render sets of devices.
  """
  data_axis_from_axis_info = {
      info: axis for axis, info in enumerate(array_axis_info)
  }

  info_by_name_or_position = {}
  has_name_only = False
  positional_count = 0
  for info in array_axis_info:
    if isinstance(info, NamedPositionalAxisInfo):
      info_by_name_or_position[info.axis_name] = info
      info_by_name_or_position[info.axis_logical_index] = info
      positional_count += 1
    elif isinstance(info, PositionalAxisInfo):
      info_by_name_or_position[info.axis_logical_index] = info
      positional_count += 1
    elif isinstance(info, NamedPositionlessAxisInfo):
      info_by_name_or_position[info.axis_name] = info
      has_name_only = True
    else:
      raise ValueError(f"Unrecognized axis info {type(info)}")

  array_shape = [info.size for info in array_axis_info]
  orig_shard_shape = sharding_info.shard_shape
  num_shards = np.prod(array_shape) // np.prod(orig_shard_shape)
  orig_device_indices_map = sharding_info.device_index_to_shard_slices
  # Possibly adjust the shard shape so that its length is the same as the array
  # shape, and so that all items in device_indices_map are slices.
  device_indices_map = {}
  shard_shape = []
  orig_shard_shape_index = 0
  first = True
  for key, ints_or_slices in orig_device_indices_map.items():
    new_slices = []
    for i, int_or_slc in enumerate(ints_or_slices):
      if isinstance(int_or_slc, int):
        new_slices.append(slice(int_or_slc, int_or_slc + 1))
        if first:
          shard_shape.append(1)
      elif isinstance(int_or_slc, slice):
        new_slices.append(int_or_slc)
        if first:
          shard_shape.append(orig_shard_shape[orig_shard_shape_index])
          orig_shard_shape_index += 1
      else:
        raise ValueError(
            f"Unrecognized axis slice in sharding info: {int_or_slc} at index"
            f" {i} for device {key}"
        )
    device_indices_map[key] = tuple(new_slices)
    first = False

  assert len(shard_shape) == len(array_shape)
  assert orig_shard_shape_index == len(orig_shard_shape)
  # Compute a truncation for visualizing a single shard. Each shard will be
  # shown as a shrunken version of the actual shard dimensions, roughly
  # proportional to the shard sizes.
  mini_trunc = arrayviz_impl.infer_balanced_truncation(
      shape=array_shape,
      maximum_size=1000,
      cutoff_size_per_axis=10,
      minimum_edge_items=2,
      doubling_bonus=5,
  )
  # Infer an axis ordering.
  known_row_infos = [info_by_name_or_position[spec] for spec in rows]
  known_column_infos = [info_by_name_or_position[spec] for spec in columns]
  row_infos, column_infos = arrayviz_impl.infer_rows_and_columns(
      all_axes=array_axis_info,
      known_rows=known_row_infos,
      known_columns=known_column_infos,
      edge_items_per_axis=mini_trunc,
  )
  # Build an actual matrix to represent each shard, with a size determined by
  # the inferred truncation.
  shard_mask = np.ones((), dtype=np.bool_)
  for t, sh_s, arr_s in zip(mini_trunc, shard_shape, array_shape):
    if t is None or sh_s <= 5:
      vec = np.ones((sh_s,), dtype=np.bool_)
    else:
      candidate = t // (arr_s // sh_s)
      if candidate <= 2:
        vec = np.array([True] * 2 + [False] + [True] * 2)
      else:
        vec = np.array([True] * candidate + [False] + [True] * candidate)
    shard_mask = shard_mask[..., None] * vec
  # Figure out which device is responsible for each shard.
  device_to_shard_offsets = {}
  shard_offsets_to_devices = collections.defaultdict(list)
  for device_index, slices in device_indices_map.items():
    shard_offsets = []
    for i, slc in enumerate(slices):
      assert slc.step is None
      if slc.start is None:
        assert slc.stop is None
        shard_offsets.append(0)
      else:
        assert slc.stop == slc.start + shard_shape[i]
        shard_offsets.append(slc.start // shard_shape[i])

    shard_offsets = tuple(shard_offsets)
    device_to_shard_offsets[device_index] = shard_offsets
    shard_offsets_to_devices[shard_offsets].append(device_index)
  # Figure out what value to show for each shard. This determines the
  # visualization color.
  shard_offset_values = {}
  shard_value_descriptions = {}
  if len(device_indices_map) <= 10 and all(
      device_index < 10 for device_index in device_indices_map.keys()
  ):
    # Map each device to an integer digit 1, 2, 3, 4, 5, 6, 7, 8, 9, 0, and
    # then draw replicas as collections of base-10 digits.
    for shard_offsets, shard_devices in shard_offsets_to_devices.items():
      if len(shard_devices) >= 7:
        # All devices are in the same shard! Arrayviz only supports 7 digits at
        # a time, so draw as much as we can.
        assert num_shards == 1
        vis_value = 1234567
      else:
        acc = 0
        for i, device_index in enumerate(shard_devices):
          acc += 10 ** (len(shard_devices) - i - 1) * (device_index + 1)
        vis_value = acc
      shard_offset_values[shard_offsets] = vis_value
      assert vis_value not in shard_value_descriptions
      shard_value_descriptions[vis_value] = (
          sharding_info.device_type
          + " "
          + ",".join(f"{d}" for d in shard_devices)
      )
    render_info_message = "Colored by device index."
  elif num_shards < 10:
    # More than ten devices, less than ten shards. Give each shard its own
    # index but start at 1.
    shard_offset_values = {
        shard_offsets: i + 1
        for i, shard_offsets in enumerate(shard_offsets_to_devices.keys())
    }
    render_info_message = "With a distinct pattern for each shard."
  else:
    # A large number of devices and shards. Start at 0.
    shard_offset_values = {
        shard_offsets: i
        for i, shard_offsets in enumerate(shard_offsets_to_devices.keys())
    }
    render_info_message = "With a distinct pattern for each shard."
  # Build the sharding visualization array.
  viz_shape = tuple(
      shard_mask.shape[i] * array_shape[i] // shard_shape[i]
      for i in range(len(array_shape))
  )
  dest = np.zeros(viz_shape, dtype=np.int32)
  destmask = np.empty(viz_shape, dtype=np.int32)
  shard_labels_by_vis_pos = [
      ["????" for _ in range(viz_shape[i])] for i in range(len(viz_shape))
  ]
  for shard_offsets, value in shard_offset_values.items():
    indexers = []
    for i, offset in enumerate(shard_offsets):
      vizslc = slice(
          offset * shard_mask.shape[i],
          (offset + 1) * shard_mask.shape[i],
          None,
      )
      indexers.append(vizslc)
      label = f"{offset * shard_shape[i]}:{(offset + 1) * shard_shape[i]}"
      for j in range(viz_shape[i])[vizslc]:
        shard_labels_by_vis_pos[i][j] = label
    dest[tuple(indexers)] = np.full_like(shard_mask, value, dtype=np.int32)
    destmask[tuple(indexers)] = shard_mask
  # Create formatting instructions to show what devices are in each shard.
  axis_lookups = [
      {
          "type": "axis_lookup",
          "axis": f"a{data_axis}",
          "skip_start": viz_shape[data_axis],
          "skip_end": viz_shape[data_axis],
          "lookup_table": {
              j: str(v)
              for j, v in enumerate(shard_labels_by_vis_pos[data_axis])
          },
      }
      for data_axis in range(len(array_shape))
  ]
  data_axis_labels = {}
  formatting_instructions = []
  formatting_instructions.append({"type": "literal", "value": "array"})

  if has_name_only:
    formatting_instructions.append({"type": "literal", "value": "[{"})

    first = True
    for data_axis, axis_info in enumerate(array_axis_info):
      if not isinstance(axis_info, NamedPositionlessAxisInfo):
        continue

      if first:
        formatting_instructions.append(
            {"type": "literal", "value": f"{repr(axis_info.axis_name)}:["}
        )
        first = False
      else:
        formatting_instructions.append(
            {"type": "literal", "value": f", {repr(axis_info.axis_name)}:["}
        )

      formatting_instructions.append(axis_lookups[data_axis])
      formatting_instructions.append({"type": "literal", "value": "]"})
      axshards = array_shape[data_axis] // shard_shape[data_axis]
      data_axis_labels[data_axis] = (
          f"{axis_info.axis_name}: {array_shape[data_axis]}/{axshards}"
      )
    formatting_instructions.append({"type": "literal", "value": "}]"})

  if positional_count:
    formatting_instructions.append({"type": "literal", "value": "["})
    for logical_index in range(positional_count):
      axis_info = info_by_name_or_position[logical_index]
      data_axis = data_axis_from_axis_info[axis_info]
      if logical_index:
        formatting_instructions.append({"type": "literal", "value": ", "})
      formatting_instructions.append(axis_lookups[data_axis])
      axshards = array_shape[data_axis] // shard_shape[data_axis]
      data_axis_labels[data_axis] = (
          f"axis {logical_index}: {array_shape[data_axis]}/{axshards}"
      )
    formatting_instructions.append({"type": "literal", "value": "]"})

  formatting_instructions.append({"type": "literal", "value": ":\n  "})
  formatting_instructions.append({
      "type": "value_lookup",
      "lookup_table": shard_value_descriptions,
      "ignore_invalid": True,
  })
  # Build the rendering.
  html_srcs = []
  html_srcs.append(
      arrayviz_impl.render_array_data_to_html(
          array_data=dest,
          valid_mask=destmask,
          column_axes=[data_axis_from_axis_info[c] for c in column_infos],
          row_axes=[data_axis_from_axis_info[r] for r in row_infos],
          slider_axes=(),
          axis_labels=[data_axis_labels[i] for i in range(len(array_shape))],
          vmin=0,
          vmax=0,
          cmap_type="digitbox",
          cmap_data=[],
          info=render_info_message,
          formatting_instructions=formatting_instructions,
          dynamic_continuous_cmap=False,
          raw_min_abs=0.0,
          raw_max_abs=0.0,
      )
  )
  html_srcs.append('<span style="font-family: monospace; white-space: pre">')
  for i, (shard_offsets, shard_devices) in enumerate(
      shard_offsets_to_devices.items()
  ):
    if i == 0:
      html_srcs.append(f"{sharding_info.device_type}")
    label = ",".join(f"{d}" for d in shard_devices)
    part = integer_digitbox(shard_offset_values[shard_offsets]).treescope_part
    assert isinstance(part, arrayviz_impl.ArrayvizDigitboxRendering)
    html_srcs.append(f"  {part.html_src} {label}")
  html_srcs.append("</span>")
  return figures_impl.TreescopeFigure(
      arrayviz_impl.ArrayvizRendering("".join(html_srcs))
  )


def render_array_sharding(
    array: ArrayInRegistry,
    rows: Sequence[int | AxisName] = (),
    columns: Sequence[int | AxisName] = (),
) -> figures_impl.TreescopeFigure:
  """Renders the sharding of an array.

  Args:
    array: The array whose sharding we should render.
    rows: Optional explicit ordering of axes for the visualization rows.
    columns: Optional explicit ordering of axes for the visualization columns.

  Returns:
    A rendering of that array's sharding.
  """
  # Retrieve the adapter for this array, which we will use to construct
  # the rendering.
  type_registries.update_registries_for_imports()
  adapter = type_registries.lookup_ndarray_adapter(array)
  if adapter is None:
    raise TypeError(
        "Cannot render sharding for array with unrecognized type"
        f" {type(array)} (not found in array adapter registry)"
    )

  # Extract information about axis names, indices, and sizes, along with the
  # sharding info.
  array_axis_info = adapter.get_axis_info_for_array_data(array)
  sharding_info = adapter.get_sharding_info_for_array_data(array)
  if sharding_info is None:
    raise ValueError(
        "Cannot render sharding for array without sharding info (not provided"
        f" by array adapter for {type(array)})."
    )

  return render_sharding_info(
      array_axis_info=array_axis_info,
      sharding_info=sharding_info,
      rows=rows,
      columns=columns,
  )


def integer_digitbox(
    value: int, *, label: str | None = None, size: str = "1em"
) -> figures_impl.TreescopeFigure:
  """Returns a "digitbox" rendering of a single integer.

  Args:
    value: Integer value to render.
    label: Optional label to draw next to the digitbox.
    size: Size for the rendering as a CSS length. "1em" means render it at the
      current font size.

  Returns:
    A renderable object showing the digitbox rendering for this integer.
  """
  value = int(value)

  render_args = json.dumps({"value": value})
  size_attr = html_escaping.escape_html_attribute(size)
  # Note: We need to save the parent of the treescope-run-here element first,
  # because it will be removed before the runSoon callback executes.
  src = (
      f'<span class="inline_digitbox" style="font-size: {size_attr}">'
      '<treescope-run-here><script type="application/octet-stream">'
      "const parent = this.parentNode;"
      "const defns = this.getRootNode().host.defns;"
      "defns.runSoon(() => {"
      f"defns.arrayviz.renderOneDigitbox(parent, {render_args});"
      "});"
      "</script></treescope-run-here>"
      "</span>"
  )
  rendering = arrayviz_impl.ArrayvizDigitboxRendering(src)
  if label:
    return figures_impl.TreescopeFigure(
        basic_parts.siblings(
            rendering,
            common_styles.custom_style(
                basic_parts.text(f" {label}"), "color:gray; font-size: 0.5em"
            ),
        )
    )
  else:
    return figures_impl.TreescopeFigure(rendering)



# Copyright 2024 The Treescope Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Embedding of external HTML content into treescope's IR."""
from __future__ import annotations

import dataclasses
import io
from typing import Any, Sequence

from treescope._internal import html_escaping
from treescope._internal.parts import part_interface

CSSStyleRule = part_interface.CSSStyleRule
JavaScriptDefn = part_interface.JavaScriptDefn
HtmlContextForSetup = part_interface.HtmlContextForSetup
RenderableTreePart = part_interface.RenderableTreePart
ExpandState = part_interface.ExpandState
FoldableTreeNode = part_interface.FoldableTreeNode


@dataclasses.dataclass(frozen=True)
class EmbeddedIFrame(RenderableTreePart):
  """Builds an HTML iframe containing scoped HTML for a rich display object.

  Attributes:
    embedded_html: HTML source to render in the iframe.
    fallback_in_text_mode: Fallback object to render in text mode.
    virtual_width: Number of characters wide to pretend this iframe is for
      layout purposes (since we can't infer it from `embedded_html` until it's
      actually rendered)
    virtual_height: Number of lines high to pretend this iframe is for layout
      purposes (since we can't infer it from `embedded_html` until it's actually
      rendered)
  """

  embedded_html: str
  fallback_in_text_mode: RenderableTreePart
  virtual_width: int = 80
  virtual_height: int = 2

  def _compute_collapsed_width(self) -> int:
    return self.virtual_width

  def _compute_newlines_in_expanded_parent(self) -> int:
    return self.virtual_height

  def foldables_in_this_part(self) -> Sequence[FoldableTreeNode]:
    return ()

  def _compute_layout_marks_in_this_part(self) -> frozenset[Any]:
    return frozenset()

  def render_to_text(
      self,
      stream: io.TextIOBase,
      *,
      expanded_parent: bool,
      indent: int,
      roundtrip_mode: bool,
      render_context: dict[Any, Any],
  ):
    # Render the fallback.
    self.fallback_in_text_mode.render_to_text(
        stream,
        expanded_parent=expanded_parent,
        indent=indent,
        roundtrip_mode=roundtrip_mode,
        render_context=render_context,
    )

  def html_setup_parts(
      self, context: HtmlContextForSetup
  ) -> set[CSSStyleRule | JavaScriptDefn]:
    rules = {
        # We want to make sure that the iframe is big enough to contain its
        # content, but this depends on how the content is formatted. Currently
        # we take the minimum of two strategies:
        # - The computed size of the content when given a frame size of 80
        #   characters wide (the initial iframe size).
        # - The size of the content scrollbar when rendered at width 0 (to catch
        #   elements that are overflowing and don't get counted by the computed
        #   size)
        # We also detect resizes of the internal content and update the size of
        # the iframe accordingly.
        JavaScriptDefn(html_escaping.without_repeated_whitespace("""
        this.getRootNode().host.defns.resize_iframe_by_content = ((iframe) => {
          const scroller = iframe.contentDocument.scrollingElement;
          scroller.style.width = "0";
          scroller.style.height = "0";
          const minWidth = scroller.scrollWidth;
          const minHeight = scroller.scrollHeight;
          scroller.style.width = "fit-content";
          scroller.style.height = "fit-content";
          scroller.style.overflow = "hidden";
          const observer = new ResizeObserver((entries) => {
            const [entry] = entries;
            const computedStyle = getComputedStyle(scroller);
            iframe.style.width =
              `calc(4ch + max(${computedStyle['width']}, ${minWidth}px))`;
            iframe.style.height =
              `calc(max(${computedStyle['height']}, ${minHeight}px))`;
          });
          observer.observe(scroller);
          console.log("registered scroller", scroller);
        });
        """)),
        CSSStyleRule(html_escaping.without_repeated_whitespace("""
          .embedded_html {
            display: block;
            padding-left: 1ch;
            padding-right: 1ch;
            width: max-content;
          }
          .embedded_html iframe {
            border: none;
            resize: both;
          }
        """)),
    }
    return rules

  def render_to_html(
      self,
      stream: io.TextIOBase,
      *,
      at_beginning_of_line: bool = False,
      render_context: dict[Any, Any],
  ):
    srcdoc = html_escaping.escape_html_attribute(
        f'<html><body style="margin: 0">{self.embedded_html}</body></html>'
    )
    stream.write(
        f'<div class="embedded_html"><iframe srcdoc="{srcdoc}"'
        ' style="width: 80ch; height: 0; overflow: hidden"'
        ' onload="this.getRootNode().host.defns.resize_iframe_by_content(this)"'
        '></iframe></div>'
    )


def embedded_iframe(
    embedded_html: str,
    fallback_in_text_mode: RenderableTreePart,
    virtual_width: int = 80,
    virtual_height: int = 2,
) -> part_interface.RenderableTreePart:
  """Returns a wrapped child in a color for non-roundtrippable abbreviations."""
  if not isinstance(embedded_html, str):
    raise ValueError(
        '`embedded_html` must be a string, but got'
        f' {type(embedded_html).__name__}'
    )
  if not isinstance(fallback_in_text_mode, RenderableTreePart):
    raise ValueError(
        '`fallback_in_text_mode` must be a renderable part, but got'
        f' {type(fallback_in_text_mode).__name__}'
    )
  if not isinstance(virtual_width, int):
    raise ValueError(
        '`virtual_width` must be an integer, but got'
        f' {type(virtual_width).__name__}'
    )
  if not isinstance(virtual_height, int):
    raise ValueError(
        '`virtual_height` must be an integer, but got'
        f' {type(virtual_height).__name__}'
    )
  return EmbeddedIFrame(
      embedded_html=embedded_html,
      fallback_in_text_mode=fallback_in_text_mode,
      virtual_width=virtual_width,
      virtual_height=virtual_height,
  )



# Copyright 2024 The Treescope Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Handler for any object with a _repr_html_.

`_repr_html_` is a method standardized by IPython, which is used to register
a pretty HTML representation of an object. This means we can support rendering
any rich object with a custom notebook rendering strategy.

See
https://ipython.readthedocs.io/en/stable/api/generated/IPython.display.html#IPython.display.display
for details on the purpose of the `_repr_html_` method.
"""

from __future__ import annotations

from typing import Any

from treescope import context
from treescope import lowering
from treescope import renderers
from treescope import rendering_parts
from treescope._internal import object_inspection


_already_processing_repr_html: context.ContextualValue[bool] = (
    context.ContextualValue(
        module=__name__,
        qualname="_already_processing_repr_html",
        initial_value=False,
    )
)
"""Tracks whether we are already rendering this subtree in some parent.

This is used to prevent situations where a node defines ``_repr_html_`` but
also includes children that define ``_repr_html_``. We only want to use the
rich HTML representation of the outermost element.
"""


def append_repr_html_when_present(
    node: Any,
    path: str | None,
    node_renderer: renderers.TreescopeSubtreeRenderer,
) -> (
    rendering_parts.RenderableTreePart
    | rendering_parts.RenderableAndLineAnnotations
    | type(NotImplemented)
):
  """Appends rich HTML representations of objects that have them."""
  if _already_processing_repr_html.get():
    # We've processed the repr_html for a parent of this node already.
    return NotImplemented

  if not isinstance(node, object_inspection.HasReprHtml):
    return NotImplemented

  # Make sure we don't try to call _repr_html_ on the children of this node,
  # since this node's _repr_html_ probably handles that already.
  with _already_processing_repr_html.set_scoped(True):
    # In "auto" mode, collapse the plaintext repr of objects with a _repr_html_,
    # and expand the HTML view instead.
    node_rendering = node_renderer(node, path=path)

  def _thunk(_):
    html_rendering = object_inspection.to_html(node)
    if html_rendering:
      return rendering_parts.embedded_iframe(
          embedded_html=html_rendering,
          fallback_in_text_mode=rendering_parts.abbreviation_color(
              rendering_parts.text("# (not shown in text mode)")
          ),
      )
    else:
      return rendering_parts.error_color(
          rendering_parts.text("# (couldn't compute HTML representation)")
      )

  iframe_rendering = lowering.maybe_defer_rendering(
      main_thunk=_thunk,
      placeholder_thunk=lambda: rendering_parts.deferred_placeholder_style(
          rendering_parts.text("...")
      ),
  )

  boxed_html_repr = rendering_parts.indented_children([
      rendering_parts.floating_annotation_with_separate_focus(
          rendering_parts.in_outlined_box(
              rendering_parts.build_full_line_with_annotations(
                  rendering_parts.build_custom_foldable_tree_node(
                      label=rendering_parts.comment_color(
                          rendering_parts.text("# Rich HTML representation")
                      ),
                      contents=rendering_parts.fold_condition(
                          expanded=iframe_rendering
                      ),
                  )
              ),
          )
      )
  ])
  return rendering_parts.siblings_with_annotations(
      node_rendering,
      rendering_parts.fold_condition(
          expanded=rendering_parts.roundtrip_condition(
              not_roundtrip=boxed_html_repr
          )
      ),
  )



# Copyright 2024 The Treescope Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""An automatic NDArray visualizer using arrayviz."""
from __future__ import annotations

import dataclasses
import sys
from typing import Any, Callable, Collection

import numpy as np
from treescope import dtype_util
from treescope import lowering
from treescope import ndarray_adapters
from treescope import rendering_parts
from treescope import type_registries
from treescope._internal import arrayviz_impl
from treescope._internal.api import arrayviz
from treescope._internal.api import autovisualize


PositionalAxisInfo = ndarray_adapters.PositionalAxisInfo
NamedPositionlessAxisInfo = ndarray_adapters.NamedPositionlessAxisInfo
NamedPositionalAxisInfo = ndarray_adapters.NamedPositionalAxisInfo
AxisInfo = ndarray_adapters.AxisInfo

ArrayInRegistry = Any


def _supported_dtype(dtype: np.dtype | None):
  return dtype is not None and (
      dtype_util.is_integer_dtype(dtype)
      or dtype_util.is_floating_dtype(dtype)
      or np.issubdtype(dtype, np.bool_)
  )


@dataclasses.dataclass
class ArrayAutovisualizer:
  """An automatic visualizer for arrays.

  ArrayAutovisualizer supports any array type registered with an NDArrayAdapter
  in the global type registry.

  Attributes:
    maximum_size: Maximum number of elements of an array to show. Arrays larger
      than this will be truncated along one or more axes.
    cutoff_size_per_axis: Maximum number of elements of each individual axis to
      show without truncation. Any axis longer than this will be truncated, with
      their visual size increasing logarithmically with the true axis size
      beyond this point.
    edge_items: How many values to keep along each axis for truncated arrays.
    prefers_column: Names that should always be assigned to columns.
    prefers_row: Names that should always be assigned to rows.
    around_zero: Whether to center continuous data around zero.
    force_continuous: Whether to always render integer arrays as continuous.
    include_repr_line_threshold: A threshold such that, if the `repr` of the
      array has fewer than that many lines, we will include that repr in the
      visualization. Useful for seeing small array values.
    token_lookup_fn: Optional function that looks up token IDs and adds them to
      the visualization on hover.
    pixels_per_cell: Size of each rendered array element in pixels, between 1
      and 21 inclusive. This controls the zoom level of the rendering. Array
      elements are always drawn at 7 pixels per cell and then rescaled, so
      out-of-bounds annotations and "digitbox" integer value patterns may not
      display correctly at fewer than 7 pixels per cell.
  """

  maximum_size: int = 4_000
  cutoff_size_per_axis: int = 128
  edge_items: int = 5
  prefers_column: Collection[str] = ()
  prefers_row: Collection[str] = ()
  around_zero: bool = True
  force_continuous: bool = False
  include_repr_line_threshold: int = 5
  token_lookup_fn: Callable[[int], str] | None = None
  pixels_per_cell: int | float = 7

  def _autovisualize_array(
      self,
      array: ArrayInRegistry,
      adapter: ndarray_adapters.NDArrayAdapter,
      path: str | None,
      label: str,
      expand_state: rendering_parts.ExpandState,
  ) -> rendering_parts.RenderableTreePart:
    """Helper to visualize an array."""
    # Extract information about axis names, indices, and sizes.
    array_axis_info = adapter.get_axis_info_for_array_data(array)

    # Assign axes, using preferred axes if possible.
    row_axes = []
    column_axes = []
    for info in array_axis_info:
      if isinstance(info, NamedPositionalAxisInfo | NamedPositionlessAxisInfo):
        if info.axis_name in self.prefers_column:
          column_axes.append(info)
        elif info.axis_name in self.prefers_row:
          row_axes.append(info)

    # Infer a good truncated shape for this array.
    edge_items_per_axis = arrayviz_impl.infer_balanced_truncation(
        tuple(info.size for info in array_axis_info),
        maximum_size=self.maximum_size,
        cutoff_size_per_axis=self.cutoff_size_per_axis,
        minimum_edge_items=self.edge_items,
    )

    row_axes, column_axes = arrayviz_impl.infer_rows_and_columns(
        all_axes=array_axis_info,
        known_rows=row_axes,
        known_columns=column_axes,
        edge_items_per_axis=edge_items_per_axis,
    )

    # Obtain truncated array and mask data from the adapter.
    truncated_array_data, truncated_mask_data = (
        adapter.get_array_data_with_truncation(
            array=array,
            mask=None,
            edge_items_per_axis=edge_items_per_axis,
        )
    )

    # Maybe infer value labels from a tokenizer.
    if (
        self.token_lookup_fn
        and not self.force_continuous
        and dtype_util.is_integer_dtype(truncated_array_data.dtype)
    ):
      tokens = np.unique(truncated_array_data.flatten()).tolist()
      value_item_labels = {
          token: self.token_lookup_fn(token) for token in tokens
      }
    else:
      value_item_labels = None

    array_rendering = arrayviz._render_pretruncated(  # pylint: disable=protected-access
        array_axis_info=array_axis_info,
        row_infos=row_axes,
        column_infos=column_axes,
        slider_infos=(),
        truncated_array_data=truncated_array_data,
        truncated_mask_data=truncated_mask_data,
        edge_items_per_axis=edge_items_per_axis,
        continuous="auto",
        around_zero=self.around_zero,
        vmax=None,
        vmin=None,
        trim_outliers=True,
        dynamic_colormap="auto",
        colormap=None,
        axis_item_labels=None,
        value_item_labels=value_item_labels,
        axis_labels=None,
        pixels_per_cell=self.pixels_per_cell,
    )
    outputs = [array_rendering]
    last_line_parts = []

    # Render the sharding as well.
    sharding_info = adapter.get_sharding_info_for_array_data(array)
    if sharding_info:
      num_devices = len(sharding_info.device_index_to_shard_slices)
      if num_devices == 1:
        [device_id] = sharding_info.device_index_to_shard_slices.keys()
        if device_id == -1:
          last_line_parts.append(f"| Device: {sharding_info.device_type}")
        else:
          last_line_parts.append(
              f"| Device: {sharding_info.device_type} {device_id}"
          )
      else:
        if sharding_info.fully_replicated:
          sharding_summary_str = (
              "Replicated across"
              f" {num_devices} {sharding_info.device_type} devices"
          )
        else:
          sharding_summary_str = (
              "Sharded across"
              f" {num_devices} {sharding_info.device_type} devices"
          )
        sharding_rendering = arrayviz.render_array_sharding(
            array,
            columns=[c.logical_key() for c in column_axes],
            rows=[r.logical_key() for r in row_axes],
        ).treescope_part
        outputs.append(
            rendering_parts.build_custom_foldable_tree_node(
                label=rendering_parts.abbreviation_color(
                    rendering_parts.siblings(
                        rendering_parts.text(sharding_summary_str),
                        rendering_parts.fold_condition(
                            expanded=rendering_parts.text(":"),
                            collapsed=rendering_parts.text(
                                " (click to expand)"
                            ),
                        ),
                    )
                ),
                contents=rendering_parts.fold_condition(
                    expanded=rendering_parts.indented_children(
                        [sharding_rendering]
                    ),
                ),
            )
        )

    # We render it with a path, but remove the copy path button. This will be
    # added back by the caller.
    if last_line_parts:
      last_line = rendering_parts.siblings(
          rendering_parts.fold_condition(
              expanded=rendering_parts.text("".join(last_line_parts)),
          ),
          rendering_parts.text(">"),
      )
    else:
      last_line = rendering_parts.text(">")
    custom_rendering = rendering_parts.build_custom_foldable_tree_node(
        label=rendering_parts.abbreviation_color(rendering_parts.text(label)),
        contents=rendering_parts.siblings(
            rendering_parts.fold_condition(
                expanded=rendering_parts.indented_children(outputs)
            ),
            rendering_parts.abbreviation_color(last_line),
        ),
        path=path,
        expand_state=expand_state,
    )
    return custom_rendering.renderable

  def __call__(
      self, value: Any, path: str | None
  ) -> autovisualize.VisualizationFromTreescopePart | None:
    """Implementation of an autovisualizer, visualizing arrays."""
    # Retrieve the adapter for this array, which we will use to construct
    # the rendering.
    adapter = type_registries.lookup_ndarray_adapter(value)
    if adapter is not None:
      if not adapter.should_autovisualize(value):
        return None

      # This is an array we can visualize!
      # Extract information about axis names, indices, and sizes.
      array_axis_info = adapter.get_axis_info_for_array_data(value)
      total_size = np.prod([ax.size for ax in array_axis_info])
      if total_size == 1:
        # Don't visualize scalars.
        return None

      np_dtype = adapter.get_numpy_dtype(value)
      if not _supported_dtype(np_dtype):
        return None

      def _placeholder() -> rendering_parts.RenderableTreePart:
        summary = adapter.get_array_summary(value, fast=True)
        return rendering_parts.deferred_placeholder_style(
            rendering_parts.text(f"<{summary}>")
        )

      def _thunk(
          expand_state: rendering_parts.ExpandState | None,
      ) -> rendering_parts.RenderableTreePart:
        # Full rendering of the array.
        if expand_state is None:
          expand_state = rendering_parts.ExpandState.WEAKLY_EXPANDED
        summary = adapter.get_array_summary(value, fast=False)
        label = f"<{summary}"
        return self._autovisualize_array(
            value, adapter, path, label, expand_state
        )

      return autovisualize.VisualizationFromTreescopePart(
          rendering_parts.RenderableAndLineAnnotations(
              renderable=lowering.maybe_defer_rendering(
                  _thunk, _placeholder, expanded_newlines_for_layout=8
              ),
              annotations=rendering_parts.build_copy_button(path),
          )
      )

    # Not an array in the registry. But it might be a JAX sharding that we can
    # visualize (if JAX is imported).
    if "jax" in sys.modules:
      import jax  # pylint: disable=import-outside-toplevel

      if isinstance(
          value,
          jax.sharding.PositionalSharding
          | jax.sharding.NamedSharding
          | jax.sharding.Mesh,
      ):
        raw_repr = repr(value)
        repr_oneline = " ".join(line.strip() for line in raw_repr.split("\n"))

        if isinstance(value, jax.sharding.PositionalSharding):
          sharding = value
          fake_axis_info = [
              PositionalAxisInfo(i, size)
              for i, size in enumerate(sharding.shape)
          ]
        elif isinstance(value, jax.sharding.NamedSharding):
          sharding = value
          # Named shardings still act on positional arrays, so show them for
          # the positional shape they require.
          fake_sizes = []
          for part in value.spec:
            if part is None:
              fake_sizes.append(1)
            elif isinstance(part, str):
              fake_sizes.append(value.mesh.shape[part])
            else:
              size = int(np.prod([value.mesh.shape[a] for a in part]))
              fake_sizes.append(size)
          fake_axis_info = [
              PositionalAxisInfo(i, size) for i, size in enumerate(fake_sizes)
          ]
        else:
          # Meshes are based on named axes. We build a temporary positional
          # sharding for visualization, but keep track of name order.
          assert isinstance(value, jax.sharding.Mesh)
          mesh = value
          sharding = jax.sharding.NamedSharding(
              value, jax.sharding.PartitionSpec(*mesh.axis_names)
          )
          fake_axis_info = [
              NamedPositionlessAxisInfo(name, mesh.shape[name])
              for name in mesh.axis_names
          ]

        fake_shape = tuple(ax.size for ax in fake_axis_info)
        some_device = next(iter(sharding.device_set))
        device_index_map = sharding.devices_indices_map(fake_shape)
        sharding_info = ndarray_adapters.ShardingInfo(
            shard_shape=sharding.shard_shape(fake_shape),
            device_index_to_shard_slices={
                d.id: v for d, v in device_index_map.items()
            },
            device_type=some_device.platform.upper(),
            fully_replicated=sharding.is_fully_replicated,
        )
        shardvis = arrayviz.render_sharding_info(
            array_axis_info=fake_axis_info,
            sharding_info=sharding_info,
        ).treescope_part
        custom_rendering = rendering_parts.build_custom_foldable_tree_node(
            label=rendering_parts.abbreviation_color(
                rendering_parts.text("<" + repr_oneline)
            ),
            contents=rendering_parts.siblings(
                rendering_parts.fold_condition(
                    expanded=rendering_parts.indented_children([shardvis])
                ),
                rendering_parts.abbreviation_color(rendering_parts.text(">")),
            ),
            path=path,
            expand_state=rendering_parts.ExpandState.EXPANDED,
        )
        return autovisualize.VisualizationFromTreescopePart(custom_rendering)
      else:
        return None

  @classmethod
  def for_tokenizer(cls, tokenizer: Any):
    """Builds an autovisualizer for a tokenizer.

    This method constructs an ArrayAutovisualizer that annotates integer array
    elements with their token strings. This can then be used to autovisualize
    tokenized arrays.

    Args:
      tokenizer: A tokenizer to use. Either a callable mapping token IDs to
        strings, or a SentencePieceProcessor.

    Returns:
      An ArrayAutovisualizer that annotates integer array elements with their
      token strings.
    """
    if callable(tokenizer):
      return cls(token_lookup_fn=lambda x: repr(tokenizer(x)))
    elif hasattr(tokenizer, "IdToPiece") and hasattr(tokenizer, "GetPieceSize"):

      def lookup(x):
        if x >= 0 and x < tokenizer.GetPieceSize():
          return repr(tokenizer.IdToPiece(x))
        else:
          return f"<out of bounds: {x}>"

      return cls(token_lookup_fn=lookup)
    else:
      raise ValueError(f"Unknown tokenizer type: {tokenizer}")



# Copyright 2024 The Treescope Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Internal definitions for figure utilities."""


from __future__ import annotations

import dataclasses

from treescope import lowering
from treescope._internal import html_escaping
from treescope._internal.parts import basic_parts
from treescope._internal.parts import part_interface


@dataclasses.dataclass(frozen=True)
class TreescopeFigure:
  """Wrapper that renders its child Treescope part as an IPython figure.

  This class implements the IPython display methods, so that it can be rendered
  to IPython at the top level.

  Attributes:
    treescope_part: Child to render.
  """

  treescope_part: part_interface.RenderableTreePart

  def _repr_html_(self) -> str:
    """Returns a rich HTML representation of this part."""
    return lowering.render_to_html_as_root(self.treescope_part, compressed=True)

  def _repr_pretty_(self, p, cycle):
    """Builds a representation of this part for the IPython prettyprinter."""
    del cycle
    p.text(lowering.render_to_text_as_root(self.treescope_part))


class InlineBlock(basic_parts.BaseSpanGroup):
  """Renders an object in "inline-block" mode."""

  def _span_css_class(self) -> str:
    return "inline_block"

  def _span_css_rule(
      self, context: part_interface.HtmlContextForSetup
  ) -> part_interface.CSSStyleRule:
    return part_interface.CSSStyleRule(
        html_escaping.without_repeated_whitespace("""
          .inline_block {
            display: inline-block;
          }
        """)
    )


class AllowWordWrap(basic_parts.BaseSpanGroup):
  """Allows line breaks in its child.."""

  def _span_css_class(self) -> str:
    return "allow_wrap"

  def _span_css_rule(
      self, context: part_interface.HtmlContextForSetup
  ) -> part_interface.CSSStyleRule:
    return part_interface.CSSStyleRule(
        html_escaping.without_repeated_whitespace("""
          .allow_wrap {
            white-space: pre-wrap;
          }
        """)
    )


class PreventWordWrap(basic_parts.BaseSpanGroup):
  """Allows line breaks in its child.."""

  def _span_css_class(self) -> str:
    return "prevent_wrap"

  def _span_css_rule(
      self, context: part_interface.HtmlContextForSetup
  ) -> part_interface.CSSStyleRule:
    return part_interface.CSSStyleRule(
        html_escaping.without_repeated_whitespace("""
          .prevent_wrap {
            white-space: pre;
          }
        """)
    )



# Copyright 2024 The Treescope Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Helpers for scoped contextual value providers."""
from __future__ import annotations

import contextlib
import dataclasses
from typing import Generic, TypeVar

from treescope._internal import docs_util

T = TypeVar("T")


@dataclasses.dataclass(frozen=True)
class _NoScopedValue:
  """Sentinel value for when no scoped value is set."""

  pass


@dataclasses.dataclass(init=False)
class ContextualValue(Generic[T]):
  """A global value which can be modified in a scoped context.

  Mutable global values can be difficult to reason about, but can also be very
  convenient for reducing boilerplate and focusing on the important parts if
  they are used carefully. Moreover, it's common to only want to change a global
  value within a particular delimited scope, without affecting anything beyond
  that scope.

  This class manages a global value that can be read anywhere, and modified
  either globally or within a delimited scope. Local values always take
  precedence over global values.

  When used within a function, it is usually recommended to use scoped updates.
  If you have many values to set, or you want to set them conditionally,
  consider using a `contextlib.ExitStack`, e.g.

  ::

      with contextlib.ExitStack() as stack:
        stack.enter_context(contextual_one.set_scoped(True))
        if foo:
          stack.enter_context(contextual_two.set_scoped(some_value))
        # logic that uses those values
      # all contexts are exited here

  or just

  ::

      stack = contextlib.ExitStack()
      stack.enter_context(contextual_one.set_scoped(True))
      if foo:
        stack.enter_context(contextual_two.set_scoped(some_value))
      # ... do something ...
      # ... then eventually:
      stack.close()

  Attributes:
    __module__: The module where this contextual value is defined.
    __qualname__: The fully-qualified name of the contextual value within its
      module.
    _raw_global_value: The current value at the global level. Should not be used
      directly.
    _raw_scoped_value: The current value at the scoped level. Should not be used
      directly.
  """

  __module__: str | None
  __qualname__: str | None
  _raw_global_value: T
  _raw_scoped_value: T | _NoScopedValue

  def __init__(
      self,
      initial_value: T,
      module: str | None = None,
      qualname: str | None = None,
  ):
    """Creates a contextual value.

    Args:
      initial_value: A value to use outside of `set_scoped` contexts.
      module: The module where this contextual value is defined. Usually this
        will be `__name__` (for whichever model it was defined in).
      qualname: The fully-qualified name of the contextual value within its
        module.
    """
    self.__module__ = module
    self.__qualname__ = qualname
    self._raw_global_value = initial_value
    self._raw_scoped_value = _NoScopedValue()

  def get(self) -> T:
    """Retrieves the current value."""
    if isinstance(self._raw_scoped_value, _NoScopedValue):
      return self._raw_global_value
    else:
      return self._raw_scoped_value

  @contextlib.contextmanager
  def set_scoped(self, new_value: T):
    # pylint: disable=g-doc-return-or-yield
    """Returns a context manager in which the value will be modified.

    This allows scoped modifications to the value, using something like

    ::

      contextual = ContextualValue(10)

      with contextual.set_scoped(3):
        v1 = contextual.get()  # v1 is 3

      v2 = contextual.get()  # v2 is 10

    If you want to conditionally set the value, consider using
    `contextlib.ExitStack`:

    ::

      with contextlib.ExitStack() as exit_stack:
        if condition:
          exit_stack.enter_context(contextual.set_scoped(3))

        # use contextual.get() here

    Args:
      new_value: The new value to use in this context.

    Returns:
      A context manager in which ``new_value`` is set.
    """
    # pylint: enable=g-doc-return-or-yield
    old_value = self._raw_scoped_value
    self._raw_scoped_value = new_value
    try:
      yield
    finally:
      assert self._raw_scoped_value is new_value
      self._raw_scoped_value = old_value

  def set_globally(self, new_value: T) -> None:
    """Sets the value at the global level.

    Updates at the global level will be overridden by any scoped updates.

    Args:
      new_value: The new value to use.
    """
    self._raw_global_value = new_value

  @docs_util.skip_automatic_documentation
  def set_interactive(self, new_value: T) -> None:
    """Alias for `set_globally`, for consistency with older Penzai API.

    Args:
      new_value: The new value to use.
    """
    self.set_globally(new_value)



# Copyright 2024 The Treescope Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Utilities for escaping HTML values."""
import re


def escape_html_attribute(attribute: str) -> str:
  """Escapes a string for rendering in a HTML attribute."""
  return attribute.replace("&", "&amp;").replace('"', "&quot;")


def without_repeated_whitespace(s: str) -> str:
  """Replaces all repeated whitespace characters with single spaces."""
  return " ".join(s.split())


def heuristic_strip_javascript_comments(js_source: str) -> str:
  """Heuristically removes javascript comments.

  Transforms a string by removing anything between "/*" and "*/", and anything
  between "//" and the next newline. This will remove JavaScript comments. Note
  that this operates directly on the unparsed string and is not guaranteed to
  preserve the semantics of the original program (for instance, it will still
  replace "comments" that appear inside string literals). However, it works
  well enough for simple JavaScript logic.

  Args:
    js_source: String to strip comments from.

  Returns:
    Version of the input string with things that look like JavaScript comments
    removed.
  """
  no_block_comments = re.sub(
      r"/\*((?!\*/).)*\*/", "", js_source, flags=re.MULTILINE | re.DOTALL
  )
  no_line_comments = re.sub(r"//[^\n]*\n", "\n", no_block_comments)
  return no_line_comments



# Copyright 2024 The Treescope Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Utilities for working with dataclasses."""

import dataclasses
import functools
import inspect
from typing import Any, TypeVar

T = TypeVar("T")


def dataclass_from_attributes(cls: type[T], **field_values) -> T:
  """Directly instantiates a dataclass given all of its fields.

  Dataclasses can override ``__init__`` to have arbitrary custom behavior, but
  this may make it difficult to construct new instances of dataclasses with
  particular field values. This function makes it possible to directly
  instantiate an instance of a dataclass with given attributes.

  Callers of this method are responsible for maintaining any invariants
  expected by the class. The intended use of this function is to restore a
  dataclass from field values extracted from another instance of that exact
  dataclass type.

  Args:
    cls: Class to instantiate.
    **field_values: Values for each of the dataclass's fields

  Returns:
    A new instance of the class.
  """
  # Make sure our fields are correct.
  expected_fields = dataclasses.fields(cls)
  expected_names = set(field.name for field in expected_fields)
  given_names = set(field_values.keys())
  if expected_names != given_names:
    raise ValueError(
        "Incorrect fields provided to `dataclass_from_attributes`; expected"
        f" {expected_names}, got {given_names}"
    )
  # Make a new object, bypassing the class's initializer.
  value = object.__new__(cls)
  # Set all the attributes, bypassing the class's __setattr__.
  for k, v in field_values.items():
    object.__setattr__(value, k, v)
  return value


@functools.cache
def init_takes_fields(cls: type[Any]) -> bool:
  """Returns True if ``cls.__init__`` takes exactly one argument per field.

  This is a heuristic for determining whether this dataclass can be rebuilt
  from its attributes using a simple repr-like format (e.g.
  ``Foo(bar=1, baz=2)``) or whether safely rebuilding it requires using
  :func:`dataclass_from_attributes` above. This is used during pretty-printing
  to determine whether to switch to a more verbose form when a round-trippable
  representation is requested.

  Note that it's technically possible to override ``__init__`` so that it takes
  the fields as attributes and then modifies them; it's not really possible to
  check for this, so we just check that the signature looks correct.

  Args:
    cls: The dataclass to check.
  """
  assert dataclasses.is_dataclass(cls)
  fields = dataclasses.fields(cls)
  remaining_field_set = set(field.name for field in fields)
  signature = inspect.signature(cls.__init__)

  # Skip `self` argument.
  parameters = list(signature.parameters.values())
  for parameter in parameters[1:]:
    if parameter.kind not in {
        inspect.Parameter.POSITIONAL_OR_KEYWORD,
        inspect.Parameter.KEYWORD_ONLY,
    }:
      # It might not be safe to pass keyword arguments.
      return False
    if parameter.name in remaining_field_set:
      remaining_field_set.remove(parameter.name)
    else:
      # Unexpected parameter; this means __init__ was overridden with extra
      # parameters.
      return False

  if remaining_field_set:
    # Some fields were not present in __init__!
    return False
  else:
    return True



# Copyright 2024 The Treescope Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Utilities for working with dtypes.

JAX extends the Numpy dtype system using the `ml_dtypes` package. Unfortunately,
these extended dtypes do not integrate directly with Numpy subdtype checks.
This module provides utilities to perform these checks that do not depend on
JAX being installed.
"""

import numpy as np
import numpy.typing


def is_integer_dtype(dtype: numpy.typing.DTypeLike) -> bool:
  """Returns whether the given dtype is an integer dtype.

  Supports both basic numpy dtypes and the extended dtypes in the `ml_dtypes`
  package (if installed).

  Args:
    dtype: The dtype to check.

  Returns:
    True if the given dtype is an integer dtype.
  """
  dtype = np.dtype(dtype)
  if np.issubdtype(dtype, np.integer):
    return True
  if isinstance(dtype.type, type) and dtype.type.__module__ == "ml_dtypes":
    import ml_dtypes  # pylint: disable=import-outside-toplevel

    try:
      _ = ml_dtypes.iinfo(dtype)
      return True
    except ValueError:
      return False
  return False


def is_floating_dtype(dtype: numpy.typing.DTypeLike) -> bool:
  """Returns whether the given dtype is a floating dtype.

  Supports both basic numpy dtypes and the extended dtypes in the `ml_dtypes`
  package (if installed).

  Args:
    dtype: The dtype to check.

  Returns:
    True if the given dtype is a floating dtype.
  """
  dtype = np.dtype(dtype)
  if np.issubdtype(dtype, np.floating):
    return True
  if isinstance(dtype.type, type) and dtype.type.__module__ == "ml_dtypes":
    import ml_dtypes  # pylint: disable=import-outside-toplevel

    try:
      _ = ml_dtypes.finfo(dtype)
      return True
    except ValueError:
      return False
  return False


def get_dtype_name(dtype: numpy.typing.DTypeLike) -> str:
  """Safely extracts a name for a dtype."""
  # Render scalar type objects as their literal names.
  if isinstance(dtype, type) and issubclass(dtype, np.generic):
    return dtype.__name__
  # Render any other dtype-like objects as the name of the concrete dtype they
  # convert to.
  try:
    return np.dtype(dtype).name
  except TypeError:
    return str(dtype)



# Copyright 2024 The Treescope Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Low-level implementation details of the foldable system.

This module contains low-level wrappers that handle folding and unfolding,
path copy button rendering, and deferred rendering.
"""

from __future__ import annotations

import dataclasses
import io
from typing import Any, Callable, Sequence

from treescope._internal import html_escaping
from treescope._internal.parts import basic_parts
from treescope._internal.parts import part_interface


CSSStyleRule = part_interface.CSSStyleRule
JavaScriptDefn = part_interface.JavaScriptDefn
HtmlContextForSetup = part_interface.HtmlContextForSetup
RenderableTreePart = part_interface.RenderableTreePart
ExpandState = part_interface.ExpandState
FoldableTreeNode = part_interface.FoldableTreeNode


SETUP_CONTEXT = HtmlContextForSetup(
    collapsed_selector=(
        ".foldable_node:has(>label>.foldable_node_toggle:not(:checked))"
    ),
    roundtrip_selector=".treescope_root.roundtrip_mode",
)

################################################################################
# Foldable node implementation
################################################################################


@dataclasses.dataclass(frozen=False)
class FoldableTreeNodeImpl(FoldableTreeNode):
  """Concrete implementation of a node that can be expanded or collapsed.

  This is kept separate from the abstract definition of a FoldableTreeNode to
  avoid strong dependencies on its implementation as much as possible.

  Attributes:
    contents: Contents of the foldable node.
    label: Optional label for the foldable node. This appears in front of the
      contents, and clicking it expands or collapses the foldable node. This
      should not contain any other foldables and should generally be a single
      line.
    expand_state: Current expand state for the node.
  """

  contents: RenderableTreePart
  label: RenderableTreePart = basic_parts.EmptyPart()
  expand_state: ExpandState = ExpandState.WEAKLY_COLLAPSED

  def get_expand_state(self) -> ExpandState:
    """Returns the node's expand state."""
    return self.expand_state

  def set_expand_state(self, expand_state: ExpandState):
    """Sets the node's expand state."""
    self.expand_state = expand_state

  def as_expanded_part(self) -> RenderableTreePart:
    """Returns the contents of this foldable when expanded."""
    return basic_parts.siblings(self.label, self.contents)

  def html_setup_parts(
      self, setup_context: HtmlContextForSetup
  ) -> set[CSSStyleRule | JavaScriptDefn]:
    if setup_context.collapsed_selector != SETUP_CONTEXT.collapsed_selector:
      raise ValueError(
          "FoldableTreeNodeImpl only works properly when the tree is"
          " configured using SETUP_CONTEXT.collapsed_selector"
      )
    # These CSS rules ensure that:
    # - Fold markers appear relative to the "foldable_node" HTML element
    # - The checkbox we are using to track collapsed state is hidden.
    # - Before the label, if this node isn't inside a collapsed parent node,
    #   we insert a floating expand/collapse marker depending on whether the
    #   checkbox is checked, and style it appropriately.
    # - If this node is in a collapsed parent node, don't toggle its state when
    #   clicked, since that won't do anything. If it is not, then show a pointer
    #   cursor to indicate that clicking will do something.
    # - When this is the first object on its line, shift the triangle marker
    #   left into the margin. Otherwise, shift the contents to the right.
    rule = html_escaping.without_repeated_whitespace(f"""
        .foldable_node
        {{
            position: relative;
        }}

        .foldable_node_toggle
        {{
            display: none;
        }}

        .foldable_node > label::before
        {{
            color: #cccccc;
            position: relative;
            left: -1ch;
            width: 0;
            display: inline-block;
        }}
        .foldable_node > label:hover::before
        {{
            color: darkseagreen;
        }}

        .foldable_node:not({setup_context.collapsed_selector} *)
          > label:has(>.foldable_node_toggle:checked)::before
        {{
            content: '\\25bc';
        }}
        .foldable_node:not({setup_context.collapsed_selector} *)
          > label:has(>.foldable_node_toggle:not(:checked))::before
        {{
            content: '\\25b6';
        }}

        .foldable_node:not({setup_context.collapsed_selector} *) > label:hover
        {{
            cursor: pointer;
        }}
        .foldable_node:is({setup_context.collapsed_selector} *) > label
        {{
            pointer-events: none;
        }}
        .foldable_node:is({setup_context.collapsed_selector} *)
        {{
            cursor: text;
        }}

        .foldable_node.is_first_on_line > label::before
        {{
            position: relative;
            left: -1.25ch;
        }}
        .foldable_node:not(
            {setup_context.collapsed_selector} *):not(.is_first_on_line)
        {{
            margin-left: 1ch;
        }}
        """)
    return (
        {CSSStyleRule(rule)}
        | self.contents.html_setup_parts(setup_context)
        | self.label.html_setup_parts(setup_context)
    )

  def render_to_html(
      self,
      stream: io.TextIOBase,
      *,
      at_beginning_of_line: bool = False,
      render_context: dict[Any, Any],
  ):
    if at_beginning_of_line:
      classname = "foldable_node is_first_on_line"
    else:
      classname = "foldable_node"

    stream.write(f'<span class="{classname}"><label>')

    stream.write('<input type="checkbox" class="foldable_node_toggle"')
    if (
        self.expand_state == ExpandState.WEAKLY_EXPANDED
        or self.expand_state == ExpandState.EXPANDED
    ):
      stream.write(" checked")
    stream.write("></input>")

    self.label.render_to_html(
        stream,
        at_beginning_of_line=at_beginning_of_line,
        render_context=render_context,
    )
    stream.write("</label>")
    self.contents.render_to_html(
        stream,
        at_beginning_of_line=isinstance(self.label, basic_parts.EmptyPart),
        render_context=render_context,
    )
    stream.write("</span>")

  def render_to_text(
      self,
      stream: io.TextIOBase,
      *,
      expanded_parent: bool,
      indent: int,
      roundtrip_mode: bool,
      render_context: dict[Any, Any],
  ):
    # In text mode, we just render the label and contents with no additional
    # wrapping. Note that we only expand if this node is expanded AND the
    # parents are expanded, since collapsed parents override children.
    expanded_here = expanded_parent and (
        self.expand_state == ExpandState.WEAKLY_EXPANDED
        or self.expand_state == ExpandState.EXPANDED
    )
    self.label.render_to_text(
        stream,
        expanded_parent=expanded_here,
        indent=indent,
        roundtrip_mode=roundtrip_mode,
        render_context=render_context,
    )
    self.contents.render_to_text(
        stream,
        expanded_parent=expanded_here,
        indent=indent,
        roundtrip_mode=roundtrip_mode,
        render_context=render_context,
    )


################################################################################
# Path copy button implementation
################################################################################


@dataclasses.dataclass
class StringCopyButton(RenderableTreePart):
  """Builds a button that, when clicked, copies the given path.

  This does nothing in text rendering mode, and only appears when its parent
  is expanded.

  Attributes:
    copy_string: String to copy when clicked
    annotation: Annotation to show when the button is hovered over.
  """

  copy_string: str
  annotation: str = "Copy: "

  def _compute_collapsed_width(self) -> int:
    return 0

  def _compute_newlines_in_expanded_parent(self) -> int:
    return 0

  def foldables_in_this_part(self) -> Sequence[FoldableTreeNode]:
    return ()

  def _compute_layout_marks_in_this_part(self) -> frozenset[Any]:
    return frozenset()

  def render_to_text(
      self,
      stream: io.TextIOBase,
      *,
      expanded_parent: bool,
      indent: int,
      roundtrip_mode: bool,
      render_context: dict[Any, Any],
  ):
    # Doesn't render at all in text mode.
    pass

  def html_setup_parts(
      self, setup_context: HtmlContextForSetup
  ) -> set[CSSStyleRule | JavaScriptDefn]:
    # pylint: disable=line-too-long
    # https://github.com/google/material-design-icons/blob/master/symbols/web/content_copy/materialsymbolsoutlined/content_copy_grad200_20px.svg
    font_data_url = "data:font/woff2;base64,d09GMgABAAAAAAIQAAoAAAAABRwAAAHFAAEAAAAAAAAAAAAAAAAAAAAAAAAAAAAABmAAgkoKdIEBCwYAATYCJAMIBCAFgxIHLhtsBMieg3FDX2LI6YvSpuiPM5T1JXIRVMvWMztPyFFC+WgkTiBD0hiDQuJEdGj0Hb/fvIdpqK6hqWiiQuMnGhHfUtAU6BNr4AFInkE6cUuun+R5qcskwvfFl/qxgEo8gbJwG81HA/nAR5LrrJ1R+gz0Rd0AJf1gN7CwGj2g0oyuR77mE16wHX9OggpeTky4eIbz5cbrOGtaAgQINwDasysQuIIXWEFwAPQpIYdU//+g7T7X3t0fKPqAv52g0LAN7AMwAmgzRS+uZSeEXx2f6czN4RHy5uBAKzBjpFp3iHQCE0ZuP4S7nfBLEHFMmAi+8vE2hn1h7+bVwXjwHrvDGUCnjfEEgt+OcZll759CJwB8h94MMGS3GZAgmI5jBQ9tTGeH9EBBIG3Dg4R/YcybAGEAAVK/AQGaAeMClAHzEOgZtg6BPgOOIDBkiQ5eFBXCBFci0phropnQAApZED1z1kSfCfthyKnHdaFsHf0NmGEN6BdAqVVpatsSZmddai92fz94Uijq6pmr6OoYCSirGmvJG3SWS3FE2cBQfT+HlopG4Fsw5agq68iZeSNlpWnBHIedMreuWqGCm1WFrkSSx526WWswAQAA"
    # pylint: enable=line-too-long

    rules = {
        JavaScriptDefn(html_escaping.without_repeated_whitespace("""
          this.getRootNode().host.defns.handle_copy_click = async (button) => {
            const dataToCopy = button.dataset.copy;
            try {
              await navigator.clipboard.writeText(dataToCopy);
              button.classList.add("was_clicked");
              setTimeout(() => {
                button.classList.remove("was_clicked");
              }, 2000);
            } catch (e) {
              button.classList.add("broken_copy");
              button.textContent = (
                  "  # Failed to copy! Copy manually instead: " + dataToCopy
              );
            }
          };
        """)),
        # Font-face definitions can't live inside a shadow
        # DOM node, so we need to inject them into the root document.
        JavaScriptDefn(html_escaping.without_repeated_whitespace("""
        if (
            !Array.from(document.fonts.values()).some(
              font => font.family == 'Material Symbols Outlined Content Copy"'
            )
        ) {
          const sheet = new CSSStyleSheet();
          sheet.replaceSync(`@font-face {
              font-family: 'Material Symbols Outlined Content Copy';
              font-style: normal;
              font-weight: 400;
              src: url({__FONT_DATA_URL__}) format('woff2');
          }`);
          document.adoptedStyleSheets = [...document.adoptedStyleSheets, sheet];
        }
        """.replace("{__FONT_DATA_URL__}", font_data_url))),
        CSSStyleRule(html_escaping.without_repeated_whitespace(f"""
          {setup_context.collapsed_selector} .copybutton {{
              display: none;
          }}
          .copybutton::before {{
              content: " ";
          }}
          .copybutton > span::before {{
              content: "\\e14d";
              font-family: 'Material Symbols Outlined Content Copy';
              -webkit-font-smoothing: antialiased;
              color: #e0e0e0;
              cursor: pointer;
              font-size: 0.9em;
          }}
          .copybutton:hover > span::before {{
              color: darkseagreen;
          }}
          .copybutton.was_clicked > span::after {{
              color: #cccccc;
          }}
          .copybutton:hover > span::after {{
              color: darkseagreen;
              content: " " var(--data-annotation) var(--data-copy);
              transition: color 1s ease-in-out;
          }}
          .copybutton.was_clicked > span::after {{
              content: " Copied! " var(--data-copy) !important;
          }}
          .copybutton.broken_copy:hover {{
              color: darkseagreen;
          }}
          .copybutton.broken_copy {{
              color: #e0e0e0;
          }}
        """)),
    }
    return rules

  def render_to_html(
      self,
      stream: io.TextIOBase,
      *,
      at_beginning_of_line: bool = False,
      render_context: dict[Any, Any],
  ):
    if self.copy_string is not None:
      copy_string_attr = html_escaping.escape_html_attribute(self.copy_string)
      copy_string_property = html_escaping.escape_html_attribute(
          repr(self.copy_string)
      )
      annotation_property = html_escaping.escape_html_attribute(
          repr(self.annotation)
      )
      attributes = html_escaping.without_repeated_whitespace(
          "class='copybutton'"
          " onClick='this.getRootNode().host.defns.handle_copy_click(this)'"
          f' data-copy="{copy_string_attr}" style="--data-copy:'
          f" {copy_string_property}; --data-annotation:"
          f' {annotation_property}" '
      )
      stream.write(f"<span {attributes}><span></span></span>")


################################################################################
# Deferrables
################################################################################


@dataclasses.dataclass(frozen=False)
class DeferredPlaceholder(basic_parts.DeferringToChild):
  """A deferred part. Renders as a placeholder which will later be replaced."""

  child: RenderableTreePart
  replacement_id: str
  saved_at_beginning_of_line: bool | None = None
  saved_render_context: dict[Any, Any] | None = None
  needs_layout_decision: bool = False

  def render_to_html(
      self,
      stream: io.TextIOBase,
      *,
      at_beginning_of_line: bool = False,
      render_context: dict[Any, Any],
  ):
    self.saved_at_beginning_of_line = at_beginning_of_line
    self.saved_render_context = render_context
    stream.write(f'<span id="{self.replacement_id}">')
    self.child.render_to_html(
        stream,
        at_beginning_of_line=at_beginning_of_line,
        render_context=render_context,
    )
    stream.write("</span>")


@dataclasses.dataclass(frozen=True)
class DeferredWithThunk:
  """Stores a deferred placeholder along with its thunk."""

  placeholder: DeferredPlaceholder
  thunk: Callable[[ExpandState | None], RenderableTreePart]


@dataclasses.dataclass(frozen=True)
class EmptyWithHeightGuess(basic_parts.BaseContentlessLeaf):
  """Helper class that reports a guess of its height."""

  fake_newlines: int

  def _compute_newlines_in_expanded_parent(self) -> int:
    return self.fake_newlines


def fake_placeholder_foldable(
    placeholder_content: RenderableTreePart, extra_newlines_guess: int
) -> FoldableTreeNode:
  """Builds a fake placeholder for deferred renderings.

  The main use for this is is to return as the placeholder for a deferred
  rendering, so that it can participate in layout decisions. The
  `get_expand_state` method can be used to infer whether the part was
  collapsed while deferred.

  Args:
    placeholder_content: The content of the placeholder to render.
    extra_newlines_guess: The number of fake newlines to pretend this object
      has.

  Returns:
    A foldable node that does not have any actual content, but pretends to
    contain the given number of newlines for layout decisions.
  """
  return FoldableTreeNodeImpl(
      contents=basic_parts.siblings(
          placeholder_content,
          EmptyWithHeightGuess(fake_newlines=extra_newlines_guess),
      ),
      expand_state=ExpandState.WEAKLY_EXPANDED,
  )



# Copyright 2024 The Treescope Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Common styles for parts of renderings."""

from __future__ import annotations

import dataclasses
import io
from typing import Any

from treescope._internal import html_escaping
from treescope._internal.parts import basic_parts
from treescope._internal.parts import part_interface


CSSStyleRule = part_interface.CSSStyleRule
JavaScriptDefn = part_interface.JavaScriptDefn
RenderableTreePart = part_interface.RenderableTreePart
HtmlContextForSetup = part_interface.HtmlContextForSetup


class AbbreviationColor(basic_parts.BaseSpanGroup):
  """Renders its child in a color for non-roundtrippable abbreviations."""

  def _span_css_class(self) -> str:
    return "color_abbrev"

  def _span_css_rule(self, context: HtmlContextForSetup) -> CSSStyleRule:
    return CSSStyleRule(html_escaping.without_repeated_whitespace("""
      .color_abbrev
      {
        color: #682a00;
      }
    """))


def abbreviation_color(
    child: part_interface.RenderableTreePart,
) -> part_interface.RenderableTreePart:
  """Returns a wrapped child in a color for non-roundtrippable abbreviations."""
  if not isinstance(child, RenderableTreePart):
    raise ValueError(
        f"`child` must be a renderable part, but got {type(child).__name__}"
    )
  return AbbreviationColor(child)


class CommentColor(basic_parts.BaseSpanGroup):
  """Renders its child in a color for comments."""

  def _span_css_class(self) -> str:
    return "color_comment"

  def _span_css_rule(self, context: HtmlContextForSetup) -> CSSStyleRule:
    return CSSStyleRule(html_escaping.without_repeated_whitespace("""
      .color_comment {
          color: #aaaaaa;
      }
      .color_comment a:not(:hover){
          color: #aaaaaa;
      }
    """))


def comment_color(
    child: part_interface.RenderableTreePart,
) -> part_interface.RenderableTreePart:
  """Returns a wrapped child in a color for comments."""
  if not isinstance(child, RenderableTreePart):
    raise ValueError(
        f"`child` must be a renderable part, but got {type(child).__name__}"
    )
  return CommentColor(child)


class ErrorColor(basic_parts.BaseSpanGroup):
  """Renders its child in red to indicate errors / problems during rendering."""

  def _span_css_class(self) -> str:
    return "color_error"

  def _span_css_rule(self, context: HtmlContextForSetup) -> CSSStyleRule:
    return CSSStyleRule(html_escaping.without_repeated_whitespace("""
      .color_error {
          color: red;
      }
    """))


def error_color(
    child: part_interface.RenderableTreePart,
) -> part_interface.RenderableTreePart:
  """Returns a wrapped child in red to indicate errors during rendering."""
  if not isinstance(child, RenderableTreePart):
    raise ValueError(
        f"`child` must be a renderable part, but got {type(child).__name__}"
    )
  return ErrorColor(child)


class DeferredPlaceholderStyle(basic_parts.BaseSpanGroup):
  """Renders its child in italics to indicate a deferred placeholder."""

  def _span_css_class(self) -> str:
    return "deferred_placeholder"

  def _span_css_rule(self, context: HtmlContextForSetup) -> CSSStyleRule:
    return CSSStyleRule(html_escaping.without_repeated_whitespace("""
      .deferred_placeholder {
          color: #a7a7a7;
          font-style: italic;
      }
    """))


def deferred_placeholder_style(
    child: part_interface.RenderableTreePart,
) -> part_interface.RenderableTreePart:
  """Returns a wrapped child in italics to indicate a deferred placeholder."""
  if not isinstance(child, RenderableTreePart):
    raise ValueError(
        f"`child` must be a renderable part, but got {type(child).__name__}"
    )
  return DeferredPlaceholderStyle(child)


class CommentColorWhenExpanded(basic_parts.BaseSpanGroup):
  """Renders its child in a color for comments, but only when expanded.

  This can be used in combination with another color (usually AbbreviationColor)
  to show a summary and then transform it into a comment when clicked, e.g. with

  AbbreviationColor(CommentColorWhenExpanded(...))
  """

  def _span_css_class(self) -> str:
    return "color_comment_when_expanded"

  def _span_css_rule(self, context: HtmlContextForSetup) -> CSSStyleRule:
    return CSSStyleRule(html_escaping.without_repeated_whitespace(f"""
      .color_comment_when_expanded:not({context.collapsed_selector} *) {{
          color: #aaaaaa;
      }}
    """))


def comment_color_when_expanded(
    child: part_interface.RenderableTreePart,
) -> part_interface.RenderableTreePart:
  """Returns a wrapped child in a color for comments, but only when expanded."""
  if not isinstance(child, RenderableTreePart):
    raise ValueError(
        f"`child` must be a renderable part, but got {type(child).__name__}"
    )
  return CommentColorWhenExpanded(child)


@dataclasses.dataclass(frozen=True)
class CustomTextColor(basic_parts.DeferringToChild):
  """A group that wraps its child in a span with a custom text color.

  Attributes:
    child: Contents of the group.
    color: CSS string defining the color.
  """

  child: RenderableTreePart
  color: str

  def render_to_html(
      self,
      stream: io.TextIOBase,
      *,
      at_beginning_of_line: bool = False,
      render_context: dict[Any, Any],
  ):
    colorstr = html_escaping.escape_html_attribute(self.color)
    stream.write(f'<span style="color:{colorstr};">')
    self.child.render_to_html(
        stream,
        at_beginning_of_line=at_beginning_of_line,
        render_context=render_context,
    )
    stream.write("</span>")


def custom_text_color(
    child: part_interface.RenderableTreePart, css_color: str
) -> part_interface.RenderableTreePart:
  """Returns a wrapped child that renders in a particular CSS color."""
  if not isinstance(child, RenderableTreePart):
    raise ValueError(
        f"`child` must be a renderable part, but got {type(child).__name__}"
    )
  if not isinstance(css_color, str):
    raise ValueError(
        f"`css_color` must be a string, but got {type(css_color).__name__}"
    )
  return CustomTextColor(child, css_color)


@dataclasses.dataclass(frozen=True)
class ColoredBorderIndentedChildren(basic_parts.IndentedChildren):
  """A sequence of children that also draws a colored line on the left.

  Assumes that the CSS variables --block-color and --block-border-scale are
  set in some container object. These are set by WithBlockColor.
  """

  def html_setup_parts(
      self, context: HtmlContextForSetup
  ) -> set[CSSStyleRule | JavaScriptDefn]:
    rule = html_escaping.without_repeated_whitespace(f"""
        .stacked_children {{
          contain: content;
        }}
        .stacked_children.colored_border
          > .stacked_child:not({context.collapsed_selector} *)
        {{
          display: block;
          margin-left: calc(2ch - 0.4ch);
        }}
        .stacked_children.colored_border:not({context.collapsed_selector} *)
        {{
          display: block;
          border-left: solid 0.4ch var(--block-color);
          padding-top: calc(0.5lh * (1 - var(--block-border-scale)));
          margin-top: calc(-0.5lh * (1 - var(--block-border-scale)));
          padding-bottom: 0.5lh;
          margin-bottom: -0.5lh;
        }}
        .stacked_children.colored_border:not({context.collapsed_selector} *)
        {{
          border-top-left-radius: calc(1em * var(--block-border-scale));
        }}
        {context.collapsed_selector} .stacked_children.colored_border
        {{
          border-top: solid 1px var(--block-color);
          border-bottom: solid 1px var(--block-color);
        }}
        """)
    return {CSSStyleRule(rule)} | super().html_setup_parts(context)

  def render_to_html(
      self,
      stream: io.TextIOBase,
      *,
      at_beginning_of_line: bool = False,
      render_context: dict[Any, Any],
  ):
    stream.write('<span class="stacked_children colored_border">')
    for part in self.children:
      # All children render at the beginning of their line.
      stream.write('<span class="stacked_child colored_border">')
      part.render_to_html(
          stream,
          at_beginning_of_line=True,
          render_context=render_context,
      )
      stream.write("</span>")

    stream.write("</span>")


def _common_block_rules(
    context: HtmlContextForSetup,
) -> dict[str, CSSStyleRule]:
  return {
      "colored_line": CSSStyleRule(html_escaping.without_repeated_whitespace("""
            .colored_line
            {
              color: black;
              background-color: var(--block-color);
              border-top: 1px solid var(--block-color);
              border-bottom: 1px solid var(--block-color);
            }
          """)),
      "patterned_line": CSSStyleRule(
          html_escaping.without_repeated_whitespace("""
            .patterned_line
            {
              color: black;
              border: 1px solid var(--block-color);
              background-image: var(--block-background-image);
              background-clip: padding-box;
            }
          """)
      ),
      "topline": CSSStyleRule(html_escaping.without_repeated_whitespace(f"""
            .topline:not({context.collapsed_selector} *)
            {{
              padding-bottom: 0.15em;
              line-height: 1.5em;
              z-index: 1;
              position: relative;
            }}
          """)),
      "bottomline": CSSStyleRule(html_escaping.without_repeated_whitespace(f"""
            .bottomline:not({context.collapsed_selector} *)
            {{
              padding-top: 0.15em;
              line-height: 1.5em;
              z-index: 1;
              position: relative;
            }}
          """)),
      "hch_space_left": CSSStyleRule(
          html_escaping.without_repeated_whitespace("""
            .hch_space_left
            {
              padding-left: 0.5ch;
            }
          """)
      ),
      "hch_space_right": CSSStyleRule(
          html_escaping.without_repeated_whitespace("""
            .hch_space_right
            {
              padding-right: 0.5ch;
            }
          """)
      ),
  }


class ColoredTopLineSpanGroup(basic_parts.BaseSpanGroup):
  """Customized span group that colors the prefix of a block."""

  def _span_css_class(self) -> str:
    return "colored_line topline hch_space_left"

  def _span_css_rule(self, context: HtmlContextForSetup) -> set[CSSStyleRule]:
    rules = _common_block_rules(context)
    return {rules["colored_line"], rules["topline"], rules["hch_space_left"]}


class ColoredSingleLineSpanGroup(basic_parts.BaseSpanGroup):
  """Customized span group that colors a single line."""

  def _span_css_class(self) -> str:
    return "colored_line hch_space_left hch_space_right"

  def _span_css_rule(self, context: HtmlContextForSetup) -> set[CSSStyleRule]:
    rules = _common_block_rules(context)
    return {
        rules["colored_line"],
        rules["hch_space_left"],
        rules["hch_space_right"],
    }


class ColoredBottomLineSpanGroup(basic_parts.BaseSpanGroup):
  """Customized span group that colors the suffix of a block."""

  def _span_css_class(self) -> str:
    return "colored_line bottomline hch_space_right"

  def _span_css_rule(self, context: HtmlContextForSetup) -> set[CSSStyleRule]:
    rules = _common_block_rules(context)
    return {
        rules["colored_line"],
        rules["bottomline"],
        rules["hch_space_right"],
    }


@dataclasses.dataclass
class WithBlockColor(basic_parts.DeferringToChild):
  """A group that configures block color CSS variables.

  This provides the necessary colors for ColoredBorderStackedChildren,
  ColoredTopLineSpanGroup, or ColoredBottomLineSpanGroup.

  Attributes:
    child: Contents of the group.
    color: CSS color to render.
  """

  child: RenderableTreePart
  color: str

  def render_to_html(
      self,
      stream: io.TextIOBase,
      *,
      at_beginning_of_line: bool = False,
      render_context: dict[Any, Any],
  ):
    style_string = f"--block-color: {self.color};"
    if at_beginning_of_line:
      style_string += "--block-border-scale: 0;"
    else:
      style_string += "--block-border-scale: 1;"
    stream.write(f'<span style="{style_string}">')
    self.child.render_to_html(
        stream,
        at_beginning_of_line=at_beginning_of_line,
        render_context=render_context,
    )
    stream.write("</span>")


class PatternedTopLineSpanGroup(basic_parts.BaseSpanGroup):
  """Customized span group that adds a pattern to the prefix of a block."""

  def _span_css_class(self) -> str:
    return "patterned_line topline hch_space_left"

  def _span_css_rule(self, context: HtmlContextForSetup) -> set[CSSStyleRule]:
    rules = _common_block_rules(context)
    return {rules["patterned_line"], rules["topline"], rules["hch_space_left"]}


class PatternedBottomLineSpanGroup(basic_parts.BaseSpanGroup):
  """Customized span group that adds a pattern to the suffix of a block."""

  def _span_css_class(self) -> str:
    return "patterned_line bottomline hch_space_left hch_space_right"

  def _span_css_rule(self, context: HtmlContextForSetup) -> set[CSSStyleRule]:
    rules = _common_block_rules(context)
    return {
        rules["patterned_line"],
        rules["bottomline"],
        rules["hch_space_left"],
        rules["hch_space_right"],
    }


class PatternedSingleLineSpanGroup(basic_parts.BaseSpanGroup):
  """Customized span group that adds a pattern to a single line."""

  def _span_css_class(self) -> str:
    return "patterned_line expand_right"

  def _span_css_rule(self, context: HtmlContextForSetup) -> set[CSSStyleRule]:
    rules = _common_block_rules(context)
    return {rules["patterned_line"], rules["expand_right"]}


@dataclasses.dataclass
class WithBlockPattern(basic_parts.DeferringToChild):
  """A group that configures block color CSS variables.

  This should be used around uses of PatternedTopLineSpanGroup,
  PatternedBottomLineSpanGroup, or PatternedSingleLineSpanGroup.

  Attributes:
    child: Contents of the group.
    color: CSS color for borders.
    pattern: CSS value for a background image pattern (e.g. using
      `repeating-linear-gradient`)
  """

  child: RenderableTreePart
  color: str
  pattern: str

  def render_to_html(
      self,
      stream: io.TextIOBase,
      *,
      at_beginning_of_line: bool = False,
      render_context: dict[Any, Any],
  ):
    style_string = (
        f"--block-color: {self.color}; "
        f"--block-background-image: {self.pattern};"
    )
    if at_beginning_of_line:
      style_string += "--block-border-scale: 0;"
    else:
      style_string += "--block-border-scale: 1;"
    stream.write(f'<span style="{style_string}">')
    self.child.render_to_html(
        stream,
        at_beginning_of_line=at_beginning_of_line,
        render_context=render_context,
    )
    stream.write("</span>")


class QualifiedTypeNameSpanGroup(basic_parts.BaseSpanGroup):
  """Customized span group that displays qualified names in a small font."""

  def _span_css_class(self) -> str:
    return "qualname_prefix"

  def _span_css_rule(self, context: HtmlContextForSetup) -> CSSStyleRule:
    return CSSStyleRule(html_escaping.without_repeated_whitespace("""
        .qualname_prefix {
          font-size: 0.8em;
        }
    """))


def qualified_type_name_style(
    child: part_interface.RenderableTreePart,
) -> part_interface.RenderableTreePart:
  """Returns a wrapped child in a small font to indicate a qualified name."""
  if not isinstance(child, RenderableTreePart):
    raise ValueError(
        f"`child` must be a renderable part, but got {type(child).__name__}"
    )
  return QualifiedTypeNameSpanGroup(child)


@dataclasses.dataclass(frozen=True)
class CSSStyled(basic_parts.DeferringToChild):
  """Adjusts the CSS style of its child.

  Attributes:
    child: Child to render.
    css: A CSS style string.
  """

  child: part_interface.RenderableTreePart
  style: str

  def render_to_html(
      self,
      stream: io.TextIOBase,
      *,
      at_beginning_of_line: bool = False,
      render_context: dict[Any, Any],
  ):
    style = html_escaping.escape_html_attribute(self.style)
    stream.write(f'<span style="{style}">')
    self.child.render_to_html(
        stream,
        at_beginning_of_line=at_beginning_of_line,
        render_context=render_context,
    )
    stream.write("</span>")


def custom_style(
    child: part_interface.RenderableTreePart,
    css_style: str,
) -> part_interface.RenderableTreePart:
  """Returns a wrapped child with a custom CSS style.

  Args:
    child: Child to render.
    css_style: A CSS style string.

  Returns:
    A wrapped child with a custom CSS style applied. Intended for an inline
    text component.
  """
  if not isinstance(child, RenderableTreePart):
    raise ValueError(
        f"`child` must be a renderable part, but got {type(child).__name__}"
    )
  if not isinstance(css_style, str):
    raise ValueError(
        f"`css_style` must be a string, but got {type(css_style).__name__}"
    )
  return CSSStyled(child, css_style)



# Copyright 2024 The Treescope Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Containers for conveniently rendering multiple objects together.

These containers make it possible to group together multiple children so that
they render as a single figure-like display object in IPython, and allow
embedding arbitrary treescope-renderable objects or treescope-compatible styles
as figures.
"""

from __future__ import annotations

import math
from typing import Any

from treescope import rendering_parts
from treescope._internal import arrayviz_impl
from treescope._internal import figures_impl
from treescope._internal import object_inspection
from treescope._internal.api import arrayviz
from treescope._internal.api import default_renderer
from treescope._internal.parts import basic_parts
from treescope._internal.parts import embedded_iframe


def inline(
    *subfigures: Any, wrap: bool = False
) -> figures_impl.TreescopeFigure:
  """Returns a figure that arranges a set of displayable objects along a line.

  Args:
    *subfigures: Subfigures to display inline.
    wrap: Whether to wrap (insert newlines) between words at the end of a line.

  Returns:
    A figure which can be rendered in IPython or used to build more complex
    figures.
  """
  siblings = rendering_parts.siblings(
      *(treescope_part_from_display_object(subfig) for subfig in subfigures)
  )
  if wrap:
    return figures_impl.TreescopeFigure(figures_impl.AllowWordWrap(siblings))
  else:
    return figures_impl.TreescopeFigure(figures_impl.PreventWordWrap(siblings))


def indented(subfigure: Any) -> figures_impl.TreescopeFigure:
  """Returns a figure object that displays a value with an indent.

  Args:
    subfigure: A value to render indented.
  """
  return figures_impl.TreescopeFigure(
      rendering_parts.indented_children([
          rendering_parts.vertical_space("0.25em"),
          treescope_part_from_display_object(subfigure),
          rendering_parts.vertical_space("0.25em"),
      ])
  )


def styled(subfigure: Any, style: str) -> figures_impl.TreescopeFigure:
  """Returns a CSS-styled version of the first figure.

  Args:
    subfigure: A value to render.
    style: A CSS style string.
  """
  return figures_impl.TreescopeFigure(
      rendering_parts.custom_style(
          treescope_part_from_display_object(subfigure), style
      )
  )


def with_font_size(
    subfigure: Any, size: str | float
) -> figures_impl.TreescopeFigure:
  """Returns a scaled version of the first figure.

  Args:
    subfigure: A value to render.
    size: A multiplier for the font size (as a float) or a string giving a
      specific CSS font size (e.g. "14pt" or "2em").
  """
  if isinstance(size, str):
    style = f"font-size: {size}"
  else:
    style = f"font-size: {size}em"
  return figures_impl.TreescopeFigure(
      rendering_parts.custom_style(
          treescope_part_from_display_object(subfigure), style
      )
  )


def with_color(subfigure: Any, color: str) -> figures_impl.TreescopeFigure:
  """Returns a colored version of the first figure.

  Args:
    subfigure: A value to render.
    color: Any CSS color string.
  """
  return figures_impl.TreescopeFigure(
      rendering_parts.custom_style(
          treescope_part_from_display_object(subfigure), f"color: {color}"
      )
  )


def bolded(subfigure: Any) -> figures_impl.TreescopeFigure:
  """Returns a bolded version of the first figure.

  Args:
    subfigure: A value to render.
  """
  return figures_impl.TreescopeFigure(
      rendering_parts.custom_style(
          treescope_part_from_display_object(subfigure), "font-weight: bold"
      )
  )


def text_on_color(
    text: str,
    value: float,
    vmax: float = 1.0,
    vmin: float | None = None,
    colormap: list[tuple[int, int, int]] | None = None,
) -> figures_impl.TreescopeFigure:
  """Renders some text on colored background, with the default colormap.

  Args:
    text: Text to display.
    value: Value to color the background with.
    vmax: Maximum value for the colormap.
    vmin: Minimum value for the colormap. Defaults to ``-vmax``.
    colormap: Explicit colormap to use. If not provided, uses the default
      diverging colormap if ``vmin`` is not provided, or the default sequential
      colormap if ``vmin`` is provided.

  Returns:
    A rendering of this word, formatted similarly to how this value would be
    formatted in an array visualization.
  """
  if colormap is None:
    if vmin is None:
      colormap = arrayviz.default_diverging_colormap.get()
    else:
      colormap = arrayviz.default_sequential_colormap.get()

  if vmin is None:
    vmin = -vmax

  if value > vmax:
    is_out_of_bounds = True
    color = colormap[-1]
  elif value < vmin:
    is_out_of_bounds = True
    color = colormap[0]
  else:
    is_out_of_bounds = False
    relative = (value - vmin) / (vmax - vmin)
    continuous_index = relative * (len(colormap) - 1)
    base_index = int(math.floor(continuous_index))
    offset = continuous_index - base_index
    if base_index == len(colormap) - 1:
      base_index -= 1
      offset = 1
    assert base_index < len(colormap) - 1
    color = [
        v0 * (1 - offset) + v1 * offset
        for v0, v1 in zip(colormap[base_index], colormap[base_index + 1])
    ]

  r, g, b = color
  # Logic matches `contrasting_style_string` in arrayviz.js
  # https://en.wikipedia.org/wiki/Grayscale#Colorimetric_(perceptual_luminance-preserving)_conversion_to_grayscale
  intensity = 0.2126 * r + 0.7152 * g + 0.0722 * b
  if intensity > 128:
    text_color = "black"
  else:
    text_color = "white"
  return figures_impl.TreescopeFigure(
      arrayviz_impl.ValueColoredTextbox(
          child=rendering_parts.text(text),
          text_color=text_color,
          background_color=f"rgb({r} {g} {b})",
          out_of_bounds=is_out_of_bounds,
          value=value,
      )
  )


def figure_from_treescope_rendering_part(
    part: rendering_parts.RenderableTreePart,
) -> figures_impl.TreescopeFigure:
  """Returns a figure object that displays a Treescope rendering part.

  Args:
    part: A Treescope rendering part to display, usually constructed via
      `repr_lib` or `rendering_parts`.

  Returns:
    A figure object that can be rendered in IPython.
  """
  return figures_impl.TreescopeFigure(part)


def treescope_part_from_display_object(
    value: Any,
) -> rendering_parts.RenderableTreePart:
  """Converts an arbitrary object to a renderable treescope part if possible.

  Behavior depends on the type of `value`:

  * If ``value`` is an instance of `TreescopeFigure`, unwraps the
    underlying treescope part.
  * If ``value`` is a string, returns a rendering of that string.
  * If ``value`` has a ``_repr_html_`` method (but isn't an instance of
    `TreescopeFigure`), returns an embedded iframe with the given HTML
    contents.
  * Otherwise, renders the value using the default treescope renderer, but
    strips off any top-level comments / copy button annotations.

  The typical use is to provide helper constructors for containers to allow
  rendering lots of different objects in the "obvious" way.

  Args:
    value: Value to wrap.

  Returns:
    A renderable treescope part showing the value.
  """
  if isinstance(value, figures_impl.TreescopeFigure):
    return value.treescope_part
  elif isinstance(value, str):
    return basic_parts.Text(value)
  else:
    maybe_html = object_inspection.to_html(value)
    if maybe_html:
      return figures_impl.InlineBlock(
          embedded_iframe.embedded_iframe(
              maybe_html,
              fallback_in_text_mode=basic_parts.Text(object.__repr__(value)),
          )
      )
    else:
      return default_renderer.build_foldable_representation(value).renderable



# Copyright 2024 The Treescope Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Stable high-level interface for building object reprs.

These functions simplify the process of implementing `__treescope_repr__` for
custom types, allowing them to integrate with treescope. This interface will be
stable across treescope releases, and may be expanded in the future to support
additional customization.
"""

from __future__ import annotations

from typing import Any, Mapping

from treescope import renderers
from treescope import rendering_parts


def render_object_constructor(
    object_type: type[Any],
    attributes: Mapping[str, Any],
    path: str | None,
    subtree_renderer: renderers.TreescopeSubtreeRenderer,
    roundtrippable: bool = False,
    color: str | None = None,
) -> rendering_parts.Rendering:
  """Renders an object in "constructor format", similar to a dataclass.

  This produces a rendering like `Foo(bar=1, baz=2)`, where Foo identifies the
  type of the object, and bar and baz are the names of the attributes of the
  object. It is a *requirement* that these are the actual attributes of the
  object, which can be accessed via `obj.bar` or similar; otherwise, the
  path renderings will break.

  This can be used from within a `__treescope_repr__` implementation via ::

    def __treescope_repr__(self, path, subtree_renderer):
      return repr_lib.render_object_constructor(
          object_type=type(self),
          attributes=<dict of attributes here>,
          path=path,
          subtree_renderer=subtree_renderer,
      )

  Args:
    object_type: The type of the object.
    attributes: The attributes of the object, which will be rendered as keyword
      arguments to the constructor.
    path: The path to the object. When `render_object_constructor` is called
      from `__treescope_repr__`, this should come from the `path` argument to
      `__treescope_repr__`.
    subtree_renderer: The renderer to use to render subtrees. When
      `render_object_constructor` is called from `__treescope_repr__`, this
      should come from the `subtree_renderer` argument to `__treescope_repr__`.
    roundtrippable: Whether evaluating the rendering as Python code will produce
      an object that is equal to the original object. This implies that the
      keyword arguments are actually the keyword arguments to the constructor,
      and not some other attributes of the object.
    color: The background color to use for the object rendering. If None, does
      not use a background color. A utility for assigning a random color based
      on a string key is given in `treescope.formatting_util`.

  Returns:
    A rendering of the object, suitable for returning from `__treescope_repr__`.
  """
  if roundtrippable:
    constructor = rendering_parts.siblings(
        rendering_parts.maybe_qualified_type_name(object_type), "("
    )
    closing_suffix = rendering_parts.text(")")
  else:
    constructor = rendering_parts.siblings(
        rendering_parts.roundtrip_condition(
            roundtrip=rendering_parts.text("<")
        ),
        rendering_parts.maybe_qualified_type_name(object_type),
        "(",
    )
    closing_suffix = rendering_parts.siblings(
        ")",
        rendering_parts.roundtrip_condition(
            roundtrip=rendering_parts.text(">")
        ),
    )

  children = []
  for i, (name, value) in enumerate(attributes.items()):
    child_path = None if path is None else f"{path}.{name}"

    if i < len(attributes) - 1:
      # Not the last child. Always show a comma, and add a space when
      # collapsed.
      comma_after = rendering_parts.siblings(
          ",",
          rendering_parts.fold_condition(collapsed=rendering_parts.text(" ")),
      )
    else:
      # Last child: only show the comma when the node is expanded.
      comma_after = rendering_parts.fold_condition(
          expanded=rendering_parts.text(",")
      )

    child_line = rendering_parts.build_full_line_with_annotations(
        rendering_parts.siblings_with_annotations(
            f"{name}=",
            subtree_renderer(value, path=child_path),
        ),
        comma_after,
    )
    children.append(child_line)

  return rendering_parts.build_foldable_tree_node_from_children(
      prefix=constructor,
      children=children,
      suffix=closing_suffix,
      path=path,
      background_color=color,
  )


def render_dictionary_wrapper(
    object_type: type[Any],
    wrapped_dict: Mapping[str, Any],
    path: str | None,
    subtree_renderer: renderers.TreescopeSubtreeRenderer,
    roundtrippable: bool = False,
    color: str | None = None,
) -> rendering_parts.Rendering:
  """Renders an object in "wrapped dictionary format".

  This produces a rendering like `Foo({"bar": 1, "baz": 2})`, where Foo
  identifies the type of the object, and "bar" and "baz" are the keys in the
  dictionary that Foo acts like. It is a *requirement* that these are accessible
  through `__getitem__`, e.g. as `obj["bar"]` or similar; otherwise, the path
  renderings will break.

  This can be used from within a `__treescope_repr__` implementation via ::

    def __treescope_repr__(self, path, subtree_renderer):
      return repr_lib.render_dictionary_wrapper(
          object_type=type(self),
          wrapped_dict=<dict of items here>,
          path=path,
          subtree_renderer=subtree_renderer,
      )

  Args:
    object_type: The type of the object.
    wrapped_dict: The dictionary that the object wraps.
    path: The path to the object. When `render_object_constructor` is called
      from `__treescope_repr__`, this should come from the `path` argument to
      `__treescope_repr__`.
    subtree_renderer: The renderer to use to render subtrees. When
      `render_object_constructor` is called from `__treescope_repr__`, this
      should come from the `subtree_renderer` argument to `__treescope_repr__`.
    roundtrippable: Whether evaluating the rendering as Python code will produce
      an object that is equal to the original object. This implies that the
      constructor for `object_type` takes a single argument, which is a
      dictionary, and that `object_type` then acts like that dictionary.
    color: The background color to use for the object rendering. If None, does
      not use a background color. A utility for assigning a random color based
      on a string key is given in `treescope.formatting_util`. (By convention,
      wrapped dictionaries aren't usually assigned a color in Treescope.)

  Returns:
    A rendering of the object, suitable for returning from `__treescope_repr__`.
  """
  if roundtrippable:
    constructor = rendering_parts.siblings(
        rendering_parts.maybe_qualified_type_name(object_type), "({"
    )
    closing_suffix = rendering_parts.text("})")
  else:
    constructor = rendering_parts.siblings(
        rendering_parts.roundtrip_condition(
            roundtrip=rendering_parts.text("<")
        ),
        rendering_parts.maybe_qualified_type_name(object_type),
        "({",
    )
    closing_suffix = rendering_parts.siblings(
        "})",
        rendering_parts.roundtrip_condition(
            roundtrip=rendering_parts.text(">")
        ),
    )

  children = []
  for i, (key, value) in enumerate(wrapped_dict.items()):
    child_path = None if path is None else f"{path}[{repr(key)}]"

    if i < len(wrapped_dict) - 1:
      # Not the last child. Always show a comma, and add a space when
      # collapsed.
      comma_after = rendering_parts.siblings(
          ",",
          rendering_parts.fold_condition(collapsed=rendering_parts.text(" ")),
      )
    else:
      # Last child: only show the comma when the node is expanded.
      comma_after = rendering_parts.fold_condition(
          expanded=rendering_parts.text(",")
      )

    key_rendering = subtree_renderer(key)
    value_rendering = subtree_renderer(value, path=child_path)

    if (
        key_rendering.renderable.collapsed_width < 40
        and not key_rendering.renderable.foldables_in_this_part()
        and (
            key_rendering.annotations is None
            or key_rendering.annotations.collapsed_width == 0
        )
    ):
      # Simple enough to render on one line.
      children.append(
          rendering_parts.siblings_with_annotations(
              key_rendering, ": ", value_rendering, comma_after
          )
      )
    else:
      # Should render on multiple lines.
      children.append(
          rendering_parts.siblings(
              rendering_parts.build_full_line_with_annotations(
                  key_rendering,
                  ":",
                  rendering_parts.fold_condition(
                      collapsed=rendering_parts.text(" ")
                  ),
              ),
              rendering_parts.indented_children([
                  rendering_parts.siblings_with_annotations(
                      value_rendering, comma_after
                  ),
                  rendering_parts.fold_condition(
                      expanded=rendering_parts.vertical_space("0.5em")
                  ),
              ]),
          )
      )

  return rendering_parts.build_foldable_tree_node_from_children(
      prefix=constructor,
      children=children,
      suffix=closing_suffix,
      path=path,
      background_color=color,
  )


def render_enumlike_item(
    object_type: type[Any],
    item_name: str,
    item_value: Any,
    path: str | None,
    subtree_renderer: renderers.TreescopeSubtreeRenderer,
) -> (
    rendering_parts.RenderableTreePart
    | rendering_parts.RenderableAndLineAnnotations
):
  """Renders a value of an enum-like type (e.g. like `enum.Enum`).

  This method can be used to render a value of a type that acts like a Python
  enum, in that there is a finite set of possible instances of the type, each of
  which have a name and a value, and where the instance can be accessed as an
  attribute (e.g. ``mytype.FOO`` is an instance of ``mytype`` with name "FOO").

  Args:
    object_type: The type of the object.
    item_name: The name of the item.
    item_value: The value of the item (``{object_type}.{item_name}.value``).
    path: The path to the object. When `render_object_constructor` is called
      from `__treescope_repr__`, this should come from the `path` argument to
      `__treescope_repr__`.
    subtree_renderer: The renderer to use to render subtrees. When
      `render_object_constructor` is called from `__treescope_repr__`, this
      should come from the `subtree_renderer` argument to `__treescope_repr__`.

  Returns:
    A rendering of the object, suitable for returning from `__treescope_repr__`.
  """
  del subtree_renderer
  return rendering_parts.build_one_line_tree_node(
      rendering_parts.siblings_with_annotations(
          rendering_parts.maybe_qualified_type_name(object_type),
          "." + item_name,
          extra_annotations=[
              rendering_parts.comment_color(
                  rendering_parts.text(f"  # value: {repr(item_value)}")
              )
          ],
      ),
      path,
  )



# Copyright 2024 The Treescope Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from absl.testing import absltest
from absl.testing import parameterized
from jax import numpy as jnp
import treescope.external.jax_support


class JaxSupportTest(parameterized.TestCase):

  @parameterized.named_parameters(
      ("bfloat16", jnp.bfloat16), ("float32", jnp.float32)
  )
  def test_compute_summary(self, dtype):
    expected = " ≈5e-05 ±2.9e-05 [≥5e-06, ≤9.5e-05] nonzero:10"
    inp = jnp.arange(10, dtype=dtype) * 1e-5 + 5e-6
    self.assertEqual(
        treescope.external.jax_support.summarize_array_data(inp), expected
    )


if __name__ == "__main__":
  absltest.main()



# Copyright 2024 The Treescope Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Handlers for basic Python types, along with namedtuples and dataclasses."""


import ast
import collections
import dataclasses
import enum
import types
from typing import Any

from treescope import formatting_util
from treescope import renderers
from treescope import rendering_parts
from treescope._internal import html_escaping
from treescope._internal.parts import basic_parts
from treescope._internal.parts import part_interface

CSSStyleRule = part_interface.CSSStyleRule
HtmlContextForSetup = part_interface.HtmlContextForSetup


class KeywordColor(basic_parts.BaseSpanGroup):
  """Renders its child in a color for keywords."""

  def _span_css_class(self) -> str:
    return "color_keyword"

  def _span_css_rule(self, context: HtmlContextForSetup) -> CSSStyleRule:
    return CSSStyleRule(html_escaping.without_repeated_whitespace("""
      .color_keyword
      {
        color: #0000ff;
      }
    """))


class NumberColor(basic_parts.BaseSpanGroup):
  """Renders its child in a color for numbers."""

  def _span_css_class(self) -> str:
    return "color_number"

  def _span_css_rule(self, context: HtmlContextForSetup) -> CSSStyleRule:
    return CSSStyleRule(html_escaping.without_repeated_whitespace("""
      .color_number
      {
        color: #098156;
      }
    """))


class StringLiteralColor(basic_parts.BaseSpanGroup):
  """Renders its child in a color for string literals."""

  def _span_css_class(self) -> str:
    return "color_string"

  def _span_css_rule(self, context: HtmlContextForSetup) -> CSSStyleRule:
    return CSSStyleRule(html_escaping.without_repeated_whitespace("""
      .color_string
      {
        color: #a31515;
      }
    """))


def render_string_or_bytes(
    node: str | bytes,
    path: str | None,
    subtree_renderer: renderers.TreescopeSubtreeRenderer,
) -> (
    rendering_parts.RenderableTreePart
    | rendering_parts.RenderableAndLineAnnotations
    | type(NotImplemented)
):
  """Renders a string or bytes literal."""
  del subtree_renderer
  lines = node.splitlines(keepends=True)
  if len(lines) > 1:
    # For multiline strings, we use two renderings:
    # - When collapsed, they render with ordinary `repr`,
    # - When expanded, they render as the implicit concatenation of per-line
    #   string literals.
    # Note that the `repr` for a string sometimes switches delimiters
    # depending on whether the string contains quotes or not, so we can't do
    # much manipulation of the strings themselves. This means that the safest
    # thing to do is to just embed two copies of the string into the IR,
    # one for the full string and the other for each line.
    return rendering_parts.build_custom_foldable_tree_node(
        contents=StringLiteralColor(
            rendering_parts.fold_condition(
                collapsed=rendering_parts.text(repr(node)),
                expanded=rendering_parts.indented_children(
                    children=[
                        rendering_parts.text(repr(line)) for line in lines
                    ],
                    comma_separated=False,
                ),
            )
        ),
        path=path,
    )
  else:
    # No newlines, so render it on a single line.
    return rendering_parts.build_one_line_tree_node(
        StringLiteralColor(rendering_parts.text(repr(node))), path
    )


def render_numeric_literal(
    node: int | float,
    path: str | None,
    subtree_renderer: renderers.TreescopeSubtreeRenderer,
) -> (
    rendering_parts.RenderableTreePart
    | rendering_parts.RenderableAndLineAnnotations
    | type(NotImplemented)
):
  """Renders a numeric literal."""
  del subtree_renderer
  return rendering_parts.build_one_line_tree_node(
      NumberColor(rendering_parts.text(repr(node))), path
  )


def render_keyword(
    node: bool | None | type(Ellipsis) | type(NotImplemented),
    path: str | None,
    subtree_renderer: renderers.TreescopeSubtreeRenderer,
) -> (
    rendering_parts.RenderableTreePart
    | rendering_parts.RenderableAndLineAnnotations
    | type(NotImplemented)
):
  """Renders a builtin constant (None, False, True, ..., NotImplemented)."""
  del subtree_renderer
  return rendering_parts.build_one_line_tree_node(
      KeywordColor(rendering_parts.text(repr(node))), path
  )


def render_enum(
    node: enum.Enum,
    path: str | None,
    subtree_renderer: renderers.TreescopeSubtreeRenderer,
) -> (
    rendering_parts.RenderableTreePart
    | rendering_parts.RenderableAndLineAnnotations
    | type(NotImplemented)
):
  """Renders an enum (roundtrippably, unlike the normal enum `repr`)."""
  del subtree_renderer
  cls = type(node)
  if node is getattr(cls, node.name):
    return rendering_parts.build_one_line_tree_node(
        rendering_parts.siblings_with_annotations(
            rendering_parts.maybe_qualified_type_name(cls),
            "." + node.name,
            extra_annotations=[
                rendering_parts.comment_color(
                    rendering_parts.text(f"  # value: {repr(node.value)}")
                )
            ],
        ),
        path,
    )
  else:
    return NotImplemented


def render_dict(
    node: dict[Any, Any],
    path: str | None,
    subtree_renderer: renderers.TreescopeSubtreeRenderer,
) -> rendering_parts.RenderableAndLineAnnotations:
  """Renders a dictionary."""

  children = []
  for i, (key, child) in enumerate(node.items()):
    if i < len(node) - 1:
      # Not the last child. Always show a comma, and add a space when
      # collapsed.
      comma_after = rendering_parts.siblings(
          ",",
          rendering_parts.fold_condition(collapsed=rendering_parts.text(" ")),
      )
    else:
      # Last child: only show the comma when the node is expanded.
      comma_after = rendering_parts.fold_condition(
          expanded=rendering_parts.text(",")
      )

    child_path = None if path is None else f"{path}[{repr(key)}]"
    # Figure out whether this key is simple enough to render inline with
    # its value.
    key_rendering = subtree_renderer(key)
    value_rendering = subtree_renderer(child, path=child_path)

    if (
        key_rendering.renderable.collapsed_width < 40
        and not key_rendering.renderable.foldables_in_this_part()
        and key_rendering.annotations.collapsed_width == 0
    ):
      # Simple enough to render on one line.
      children.append(
          rendering_parts.siblings_with_annotations(
              key_rendering, ": ", value_rendering, comma_after
          )
      )
    else:
      # Should render on multiple lines.
      children.append(
          rendering_parts.siblings(
              rendering_parts.build_full_line_with_annotations(
                  key_rendering,
                  ":",
                  rendering_parts.fold_condition(
                      collapsed=rendering_parts.text(" ")
                  ),
              ),
              rendering_parts.indented_children([
                  rendering_parts.siblings_with_annotations(
                      value_rendering, comma_after
                  ),
                  rendering_parts.fold_condition(
                      expanded=rendering_parts.vertical_space("0.5em")
                  ),
              ]),
          )
      )

  if type(node) is dict:  # pylint: disable=unidiomatic-typecheck
    start = "{"
    end = "}"
  else:
    start = rendering_parts.siblings(
        rendering_parts.maybe_qualified_type_name(type(node)), "({"
    )
    end = "})"

  if not children:
    return rendering_parts.build_one_line_tree_node(
        line=rendering_parts.siblings(start, end), path=path
    )
  else:
    return rendering_parts.build_foldable_tree_node_from_children(
        prefix=start,
        children=children,
        suffix=end,
        path=path,
    )


def render_sequence_or_set(
    sequence: dict[Any, Any],
    path: str | None,
    subtree_renderer: renderers.TreescopeSubtreeRenderer,
) -> rendering_parts.RenderableAndLineAnnotations:
  """Renders a sequence or set to a foldable."""
  if (
      isinstance(sequence, tuple)
      and type(sequence) is not tuple  # pylint: disable=unidiomatic-typecheck
      and hasattr(type(sequence), "_fields")
  ):
    # This is actually a namedtuple, which renders with keyword arguments.
    return render_namedtuple_or_ast(sequence, path, subtree_renderer)

  children = []
  for i, child in enumerate(sequence):
    child_path = None if path is None else f"{path}[{repr(i)}]"
    children.append(subtree_renderer(child, path=child_path))

  force_trailing_comma = False
  if isinstance(sequence, tuple):
    before = "("
    after = ")"
    if type(sequence) is not tuple:  # pylint: disable=unidiomatic-typecheck
      # Subclass of `tuple`.
      assert not hasattr(type(sequence), "_fields"), "impossible: checked above"
      # Unusual situation: this is a subclass of `tuple`, but it isn't a
      # namedtuple. Assume we can call it with a single ordinary tuple as an
      # argument.
      before = rendering_parts.siblings(
          rendering_parts.maybe_qualified_type_name(type(sequence)),
          "(" + before,
      )
      after = after + ")"
    force_trailing_comma = len(sequence) == 1
  elif isinstance(sequence, list):
    before = "["
    after = "]"
    if type(sequence) is not list:  # pylint: disable=unidiomatic-typecheck
      before = rendering_parts.siblings(
          rendering_parts.maybe_qualified_type_name(type(sequence)),
          "(" + before,
      )
      after = after + ")"
  elif isinstance(sequence, set):
    if not sequence:
      before = "set("
      after = ")"
    else:  # pylint: disable=unidiomatic-typecheck
      before = "{"
      after = "}"

    if type(sequence) is not set:  # pylint: disable=unidiomatic-typecheck
      before = rendering_parts.siblings(
          rendering_parts.maybe_qualified_type_name(type(sequence)),
          "(" + before,
      )
      after = after + ")"
  elif isinstance(sequence, frozenset):
    before = "frozenset({"
    after = "})"
    if type(sequence) is not frozenset:  # pylint: disable=unidiomatic-typecheck
      before = rendering_parts.siblings(
          rendering_parts.maybe_qualified_type_name(type(sequence)),
          "(" + before,
      )
      after = after + ")"
  else:
    raise ValueError(f"Unrecognized sequence {sequence}")

  if not children:
    return rendering_parts.build_one_line_tree_node(
        line=rendering_parts.siblings(before, after), path=path
    )
  else:
    return rendering_parts.build_foldable_tree_node_from_children(
        prefix=before,
        children=children,
        suffix=after,
        path=path,
        comma_separated=True,
        force_trailing_comma=force_trailing_comma,
    )


def render_simplenamespace(
    node: types.SimpleNamespace,
    path: str | None,
    subtree_renderer: renderers.TreescopeSubtreeRenderer,
) -> rendering_parts.RenderableAndLineAnnotations:
  """Renders a SimpleNamespace."""
  return rendering_parts.build_foldable_tree_node_from_children(
      prefix=rendering_parts.siblings(
          rendering_parts.maybe_qualified_type_name(type(node)), "("
      ),
      children=rendering_parts.build_field_children(
          node,
          path,
          subtree_renderer,
          fields_or_attribute_names=tuple(node.__dict__.keys()),
      ),
      suffix=")",
      path=path,
  )


def render_namedtuple_or_ast(
    node: tuple[Any, ...] | ast.AST,
    path: str | None,
    subtree_renderer: renderers.TreescopeSubtreeRenderer,
) -> rendering_parts.RenderableAndLineAnnotations:
  """Renders a namedtuple or AST class."""
  ty = type(node)
  assert hasattr(ty, "_fields")
  return rendering_parts.build_foldable_tree_node_from_children(
      prefix=rendering_parts.siblings(
          rendering_parts.maybe_qualified_type_name(ty), "("
      ),
      children=rendering_parts.build_field_children(
          node, path, subtree_renderer, fields_or_attribute_names=ty._fields
      ),
      suffix=")",
      path=path,
  )


BUILTINS_REGISTRY = {
    # Builtin atomic types.
    str: render_string_or_bytes,
    bytes: render_string_or_bytes,
    int: render_numeric_literal,
    float: render_numeric_literal,
    bool: render_keyword,
    type(None): render_keyword,
    type(NotImplemented): render_keyword,
    type(Ellipsis): render_keyword,
    enum.Enum: render_enum,
    # Builtin basic structures.
    dict: render_dict,
    tuple: render_sequence_or_set,
    list: render_sequence_or_set,
    set: render_sequence_or_set,
    frozenset: render_sequence_or_set,
    types.SimpleNamespace: render_simplenamespace,
    ast.AST: render_namedtuple_or_ast,
    collections.UserDict: render_dict,
}


def handle_basic_types(
    node: Any,
    path: str | None,
    subtree_renderer: renderers.TreescopeSubtreeRenderer,
) -> (
    rendering_parts.RenderableTreePart
    | rendering_parts.RenderableAndLineAnnotations
    | type(NotImplemented)
):
  """Renders basic builtin Python types."""
  candidate_type = type(node)
  for supertype in candidate_type.__mro__:
    if supertype in BUILTINS_REGISTRY:
      return BUILTINS_REGISTRY[supertype](node, path, subtree_renderer)

  if dataclasses.is_dataclass(node) and not isinstance(node, type):
    constructor_open = rendering_parts.render_dataclass_constructor(node)
    if hasattr(node, "__treescope_color__") and callable(
        node.__treescope_color__
    ):
      background_color, background_pattern = (
          formatting_util.parse_simple_color_and_pattern_spec(
              node.__treescope_color__(), type(node).__name__
          )
      )
    else:
      background_color = None
      background_pattern = None

    return rendering_parts.build_foldable_tree_node_from_children(
        prefix=constructor_open,
        children=rendering_parts.build_field_children(
            node,
            path,
            subtree_renderer,
            fields_or_attribute_names=dataclasses.fields(node),
        ),
        suffix=")",
        path=path,
        background_color=background_color,
        background_pattern=background_pattern,
    )

  return NotImplemented
